(**
 dns.ml --- OCaml DNS library
                                                                         
 This file is part of the ODNS library.

 Copyright (C) 2002  Guillaume Valadon <guillaume@valadon.net>
                     Manu Letellier <manu.letellier@gmail.com>
               2010-2011  Jehan Hysseo <hysseo at zemarmot point net>
                                                                         
 This program is free software; you can redistribute it and/or modify it 
 under the terms of the GNU General Public License version 2 as          
 published by the Free Software Foundation.                              
                                                                         
 This program is distributed in the hope that it will be useful, but 
 WITHOUT ANY WARRANTY; without even the implied warranty of            
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     
 General Public License for more details.                                

 @see <http://tools.ietf.org/html/rfc1035> Domain names - Implementation and specification
 @see <http://tools.ietf.org/html/rfc2782> DNS SRV
 @see <http://www.ietf.org/rfc/rfc3596.txt> DNS Extensions to Support IP Version 6
 @author Guillaume Valadon (original, version 0.1)
 @author Jehan Hysseo (current author and maintainer, after version 0.1)
 @version 0.2

*)

exception Unknown_type of string;;

exception Unknown_class of string;;

(* TimedOut is an internal exception, it won't "get out". *)
exception TimedOut;;

(* TODO: apparently the Malformed exception is caught nowhere. This should be
 * fixed as it is not acceptable that a malformed return from a DNS server (or
 * any kind of other server if for instance this is not a DNS server which
 * listens at port 53) crashes a program. *)
exception Malformed of string;;

(** Get an {i integer} that represents the ressource record type from 
    an {i string}. *)
let string_to_qtype str =
   match str with
   | "A" -> 1
   | "NS" -> 2
   | "MD" ->
       Printf.eprintf "%s is an obsolete_type, it is adviced to use MX \
       instead.\n" str;
       3
   | "MF" ->
       Printf.eprintf "%s is an obsolete_type, it is adviced to use MX \
       instead.\n" str;
       4
   | "CNAME" -> 5
   | "SOA" -> 6
   | "MB" -> 7
   | "MG" -> 8
   | "MR" -> 9
   | "NULL" -> 10
   | "WKS" -> 11
   | "PTR" -> 12 
   | "HINFO" -> 13
   | "MINFO" -> 14
   | "MX" -> 15
   | "TXT" -> 16
   | "AXFR" -> 252
   | "MAILB" -> 253
   | "MAILA" ->
       (*Printf.eprintf "%s is an obsolete_type, it is adviced to use MX \
       instead.\n" str;*)
       254
   | "ANY" | "*" -> 255
   | "AAAA" -> 28
   | "SRV" -> 33
   | _ -> raise (Unknown_type str)
;;

let string_to_qclass = function
    | "IN" -> 1
    | "CS" ->
            (* TODO: make it an output only in -verbose mode. *)
            prerr_string "the CSNET class is obsolete (used only for examples in some \
                obsolete RFCs)";
            2
    | "CH" -> 3 (* TODO: Unsupported_class, no? *)
    | "HS" -> 4
    | "*" -> 255
    | c -> raise (Unknown_class c);;

(** {1 DNS message construction} *)

(** Here are generic classes with methods to easily construct DNS message from the roots. 
 Some knowledge of the DNS is required here. You might probably want to look the
 next section for higher level DNS query constructors. *)

(** Type that represents the data type, only internet data (IN) is 
    really useful. *)
type classtype = 
   | IN (* 1 *)
   | CS (* 2 *)
   | CH (* 3 *)
   | HS (* 4 *)
;;

(** QR field of message header. *)
type message_type =
    | Query
    | Response;;

(** Type that represents resource records. *)
type rdata = 
   | A of string
   | NS of string
   | MD of string (* Obsolete - use MX *)
   | MF of string (* Obsolete - use MX *) 
   | CNAME of string
   | SOA of string * string * int * int * int * int * int
   | MB of string
   | MG of string
   | MR of string
   | NULL of string (* Anything ... *)
   | WKS of string * int * (int list)
   | PTR of string
   | HINFO of string * string
   | MINFO of string * string
   | MX of int * string
   | TXT of string list
   | SRV of int * int * int * string (* Priority, weight, port *)
   | AAAA of string
   | Unsupported_RDATA of int * string (* Unsupported type, raw rdata field *)
;;

(** Give an {i integer} from a {i Dns.rrtype}. *)
let rrtype_to_int rr = 
   match rr with
   | A _-> 1
   | NS _ -> 2
   | MD _ -> 3
   | MF _ -> 4
   | CNAME _ -> 5
   | SOA _ -> 6
   | MB _ -> 7
   | MG _ -> 8
   | MR _ -> 9
   | NULL _ -> 10
   | WKS _ -> 11
   | PTR _ -> 12 
   | HINFO _ -> 13
   | MINFO _ -> 14
   | MX _ -> 15
   | TXT _ -> 16
   | AAAA _-> 28 
   | SRV _ -> 33
   | Unsupported_RDATA (t, _) -> t
;;

(** Node of a N-tree, that represents the dns name space. *)
type dnstree = DomainName of string * dnstree list * rdata list;;

(** Construct a 'dns tree' easily. *)
class dns_auth =
object (self)

   val mutable root = []
  
   (** Set the new root of the dns tree. *)
   method set aroot = root <- aroot

   (** Clear the dns tree. *)
   method clear = root <- []
   
   (** Get {i str} resource record. *)
   method get (str:string) =
      let rec temp root strlst =
      match root with
      | DomainName (str, domainlst, rrlst)::l
      when [str] = strlst ->
	 rrlst
      | DomainName (str, domainlst, rrlst)::l
      when str = (List.hd strlst) ->
	 temp domainlst (List.tl strlst)
      | DomainName (_,_,_)::l ->
	 temp l strlst
      | [] -> 
          failwith "TODO"
          (*[NotFound]*)
      in
      temp root (self#str2list str)

   (** Get {i str} subtree (very useful for zone transfer). *)
   method get_subtree (str:string) =
      let rec temp root strlst =
      match root with
      | DomainName (str, domainlst, rrlst)::l
      when [str] = strlst ->
	 [DomainName (str, domainlst, rrlst)]
      | DomainName (str, domainlst, rrlst)::l
      when str = (List.hd strlst) ->
	 [DomainName (str, temp domainlst (List.tl strlst), rrlst)]
      | DomainName (_,_,_)::l ->
	 temp l strlst
      | [] -> 
	 []
      in
      temp root (self#str2list str)

   (** Add an entry and its resource record. *)
   method add str (rrlist: rdata list) =
      match self#get str with
      (*| [NotFound] -> self#replace str rrlist*) (*TODO*)
      | lst -> self#replace str (lst@rrlist)

   (** Replace resource record of this entry. *)
   method replace str (rrlist: rdata list) =
      let rec temp root strlst rrlst =
      let hd = List.hd strlst
      and tl = List.tl strlst in
      match root with
      | [] ->
	 (match tl with
	 | [] ->
	       [DomainName (hd, [], rrlst)]
	 | _ ->
	       [DomainName (hd, temp [] tl rrlst, [])])
      | lst ->
	 (let test = ref true
   	 and recnull strlst a b c rrlst=
	    if strlst = []
	    then
	       DomainName (a, b, rrlst)
	    else
	       DomainName (a, temp b tl rrlst, c)
	 in
	 let tlst = List.map 
	    (fun (DomainName (a, b, c)) ->
	       match a = hd with
	       | false -> DomainName (a,b,c)
	       | true -> test := false; recnull tl a b c rrlst
	       ) lst in
	 match !test with
	 | false ->
	    tlst
	 | true ->
	    lst@(temp [] strlst rrlst))
      in 
      root <- temp root (self#str2list (String.lowercase str)) rrlist

   (** Check if we have authoritative data for the domain {i str}. *)
   method is_authoritative (str: string) = 
      let rec temp root strlst =
      match root with
      | DomainName (str, domainlst, rrlst)::l ->
         let hd = List.hd strlst 
         and tl = List.tl strlst in
         (match str = hd with
         | true ->
            if tl = []
            then 
               true
            else
               true&&(temp domainlst tl)
         | false -> temp l strlst)
      | _ -> false
                    
      in
      temp root (self#str2list (String.lowercase str))
   
   (** Extract domain from {i str}. *)
   method get_domain str =
      try
      let _ = Str.search_forward (Str.regexp "\\.\\(.*\\)$") str 0 in 
      Str.matched_group 1 str;
      with
      | Not_found -> ""

    method private build_list lst =
	let rec temp lst =
	match lst with
	| [] -> ""
	| a::l -> (let test = temp l in
		  if test = ""
		  then
		    a
		  else
		    a^test
		  )
	and temp2 lst =
	match lst with
	| [] -> []
	| a::l -> (let test = temp l in
		  if test = ""
		  then
		    [a]@(temp2 l)
		  else
		    [a^"."^test]@(temp2 l)
		  )
	in
	temp2 lst

   method private str2list str =
     List.rev (Str.split (Str.regexp "\\.") str)

end;;


(** Represents a dns header. *)
class header =
object (self)
   initializer Random.self_init ()

   (* Each line in its default value. *)
   val mutable id = Random.int 65535
   val mutable flags = 256
   val mutable qdcount = 0
   val mutable ancount = 0
   val mutable nscount = 0
   val mutable arcount = 0

    (* TODO: remove. *)
   method flags = flags
   method set_flags aflags = flags <- aflags land 0xFFFF

   (* ID *)
   method id = id
   method set_id aid = id <- aid land 0xFFFF (* TODO: shouldn't I rather rather an exception? *)

   (* QR bit *)
   method qr = 
       if (flags land 0x8000) = 0x8000
       then
           Response
       else
           Query

   method set_qr qr_flag =
     match qr_flag with 
     | Query ->
             flags <- flags land 0x7FFF
     | Response ->
             flags <- flags lor 0x8000

   (* RA (recursion available bit *)
   method is_ra = (flags land 0x0080) = 0x0080

   method set_ra boolean =
     let oflags = flags land 0xFF7F in
     flags <- if boolean then oflags lor 0x0080 else flags

   (* Z bits (private because reserved for future use. Now be always be 0) *)  
   method z = flags land 0x0070

   method set_z num =
      flags <- (flags land 0xFF8F) lor ((num land 0xF) lsl 0x4)

  (* QDCOUNT *)
   method set_qdcount aqdcount = qdcount <- aqdcount land 0xFFFF
   method qdcount = qdcount

   (* ANCOUNT *)
   method set_ancount aancount = ancount <- aancount land 0xFFFF
   method ancount = ancount

   (* NSCOUNT *)
   method set_nscount anauth = nscount <- anauth land 0xFFFF
   method nscount = nscount

   (* ARCOUNT *)
   method arcount = arcount
   method set_arcount anaddr = arcount <- anaddr land 0xFFFF

   (* set flags methods *)
   (* OPCODE 4 bits *)
   method opcode = (flags land 0x7800) lsr 11
   method set_opcode num =
      flags <- (flags land 0x87FF) lor ((num land 0xF) lsl 0xB)
      (* TODO: should I make more human readable opcode type? *)

   (* AA bit *)   
   method is_aa = (flags land 0x0400) = 0x0400
   method set_aa boolean =
     let oflags = flags land 0xFBFF in
     flags <- if boolean then oflags lor 0x0400 else flags

   (* TC bit *)
   method is_tc = (flags land 0x0200) = 0x0200
   method set_tc boolean =
     let oflags = flags land 0xFDFF in
     flags <- if boolean then oflags lor 0x0200 else flags

   (* RD bit *)
   method is_rd = (flags land 0x0100) = 0x0100
   method set_rd boolean = 
     let oflags = flags land 0xFEFF in
     flags <- if boolean then oflags lor 0x0100 else flags

   (* RCODE bits *) 
   method rcode = flags land 0xF
   method set_rcode num = 
      flags <- (flags land 0xFFF0) lor (num land 0xF)
      (* TODO: should I make a more human readable rcode type? *)
  

   (** Output the header's content as a string. *)
   method output =
   (Tools.hexa_of_int id 2)^(Tools.hexa_of_int flags 2)^
   (Tools.hexa_of_int qdcount 2)^(Tools.hexa_of_int ancount 2)^
   (Tools.hexa_of_int nscount 2)^(Tools.hexa_of_int arcount 2)

end;;

(** Represents a question section of a DNS message. *)
class question =
object (self)

    val mutable qname = "" 
    val mutable qtype = 0
    val mutable qclass = 0
    
    method qname =
        qname

    method set_qname query_name =
        qname <- (query_name
        ^
        (* A domain name ends by a dot in RFC way. *)
        if query_name = "" || query_name.[(String.length query_name) - 1] = '.'
        then
           "" 
        else
            ".")

    method qtype = (*Dnsprint.type_to_string*) qtype
    method set_qtype query_type = qtype <- query_type land 0xFFFF
    (* It is more useable to set the type as a string than as its type id. *)

    method qclass = (*Dnsprint.class_to_string*) qclass
    method set_qclass query_class = qclass <- query_class land 0xFFFF

    (** Output the dns question's content as a string. *)
    method output =
    (Tools.format_domain_name qname)
    ^ (Tools.hexa_of_int qtype 2)
    ^ (Tools.hexa_of_int qclass 2)

end;;

(** Represents a dns Resource Record. *)
class rr =
object (self)
  (*inherit question as super*)

    val mutable name = "" 
    val mutable _type = 0
    val mutable _class = 0
    
    method name = name
    method set_name aname = name <- aname

    method _type = (*Dnsprint.type_to_string*) _type
    (* TODO: it is more useable to set the type as a string than as its type id. *)

    method _class = (*Dnsprint.class_to_string*) _class
    method set_class query_class = _class <- query_class land 0xFFFF

   val mutable ttl = 0
   val mutable rdlength = 0
   val mutable rdata = A ""

   method ttl = ttl
   method set_ttl attl = ttl <- attl land 0xFFFF

   method rdlength =
       rdlength

   method rdata = rdata
   method set_rdata new_rdata =
     (* TODO: for known rdata, should I not compress with label pointers? *)
       rdata <- new_rdata;
       self#compute_rdlength;
       _type <- rrtype_to_int rdata

   method private output_rdata =
      match rdata with
      | A adr ->
          Tools.format_address adr
      | NS nsdname ->
          Tools.format_domain_name nsdname
      | MD dom ->
          Tools.format_domain_name dom 
      | MF dom ->
          Tools.format_domain_name dom 
      | CNAME dom ->
          Tools.format_domain_name dom 
      | SOA (mname, rname, serial, refresh, retry, expire, min) ->
          (Tools.format_domain_name mname)
          ^ (Tools.format_domain_name rname)
          ^ (Tools.hexa_of_int serial 4)
          ^ (Tools.hexa_of_int refresh 4)
          ^ (Tools.hexa_of_int retry 4)
          ^ (Tools.hexa_of_int expire 4)
          ^ (Tools.hexa_of_int min 4)
      | MB dom ->
          Tools.format_domain_name dom 
      | MG dom ->
          Tools.format_domain_name dom 
      | MR dom ->
          Tools.format_domain_name dom 
      | NULL str ->
          str
      | WKS (adr, protocol, ports) ->
          Tools.format_address adr
          ^ (Tools.hexa_of_int protocol 1)
          ^
          (try
              (* I want the bigger port first to compute the bitmap's size. *)
              let ports = List.fast_sort (fun a b -> - compare a b) ports in
              let bitmap_size = 1 + (List.hd ports) / 8 in
              let bitmap = String.make bitmap_size (Char.chr 0) in
              let rec fill = function
                  | [] -> bitmap
                  | p::l ->
                          let c = bitmap.[p / 8] in
                          let bit = 1 lsl (p mod 8) in 
                          let c2 = Char.chr ((Char.code c) lor bit) in
                          bitmap.[p / 8] <- c2;
                          fill l
              in fill ports
          with
          | Failure "hd" -> ""
          (* Is it ok if the bitmap is null? Or should it be at least 8 bits
           * long? In the RFC, there is no obvious interdiction of a null
           * length. *)
              )
      | PTR ptrdname ->
          Tools.format_domain_name ptrdname
      | HINFO (cpu, os) ->
            (Tools.format_character_string cpu)
            ^ (Tools.format_character_string os)
      | MINFO (rmailbx, emailbx) ->
          Tools.format_domain_name rmailbx
          ^ Tools.format_domain_name emailbx 
      | MX (pref, name) ->
          (Tools.hexa_of_int pref 2)
          ^ (Tools.format_domain_name name)
      | TXT strings ->
          List.fold_left (^) "" (List.map Tools.format_character_string strings)
      | SRV (prio, weight, port, target) -> (Tools.hexa_of_int prio
      4)^(Tools.hexa_of_int weight 4)^(Tools.hexa_of_int port 4)^(Tools.format_domain_name target)
      | _ -> "#output_rdata: Error"

   method private compute_rdlength =
      let temp =
      match (self#rdata) with
      | A _ -> (* OK *)
          4
      | MD dom | MF dom | MB dom | MG dom | MR dom
      | NS dom | PTR dom
      | CNAME dom -> (* OK *)
          String.length (Tools.format_domain_name dom)
      | AAAA _ -> (* OK *)
          16
      | HINFO (cpu,os) -> (* OK *)
          String.length cpu
          + String.length os
          + 2
      | MINFO (rmailbx, emailbx) -> (* OK *)
          String.length (Tools.format_domain_name rmailbx)
          + String.length (Tools.format_domain_name emailbx)
      | MX (_, exchange) -> (* OK *)
          2 +
          String.length (Tools.format_domain_name exchange)
      | NULL str -> (* OK *)
          String.length str
      | WKS (adr, protocol, ports) -> (* OK *)
          5 +
          begin
              match ports with
              | [] -> 0
              | p ->
                  (* I want the bigger port first to compute the bitmap's size. *)
                  let p = List.fast_sort (fun a b -> - compare a b) p in
                  (* I know this List.hd will not raise an exception as I
                   * checked it was not empty. *)
                  1 + (List.hd p) / 8
          end
      | SOA (m,r,s,ref,ret,ex,min) -> (* OK *)
          String.length (Tools.format_domain_name m) +
          String.length (Tools.format_domain_name r) + 5 * 4
      | TXT strings -> (* OK *)
          List.fold_left (+) (List.length strings) (List.map String.length strings)
      | SRV (prio, weight, port, target) -> (* OK *)
          6 + String.length (Tools.format_domain_name target)
      | Unsupported_RDATA (_, raw_rdata) -> (* OK *)
          String.length raw_rdata
      in
      rdlength <- temp

   (** Output the dns response's content as a string. *)
   method output = 
    (Tools.format_domain_name name)
    ^ (Tools.hexa_of_int _type 2)
    ^ (Tools.hexa_of_int _class 2)
    ^ (Tools.hexa_of_int ttl 4)
    ^ (Tools.hexa_of_int rdlength 2)
    ^ self#output_rdata

end;;

(** Represents a dns packet. *)
class message =
object (self)
   
   val mutable header = new header;
   val mutable questions = ([]: question list)
   val mutable responses = ([]: rr list)
   val mutable authorities = ([]: rr list)
   val mutable additionals = ([]: rr list)

   method header = header
   method questions = questions
   method answers = responses
   method authorities = authorities
   method additionals = additionals

   method add_questions oq =
       questions <- questions@oq;
       header#set_qdcount (List.length questions)

   method set_questions qs =
       questions <- qs

   method add_answers ore =
       responses <- responses@ore;
       header#set_ancount (List.length responses)

   method set_answers a =
       responses <- a

   method add_authorities oau =
       authorities <- authorities@oau;
       header#set_nscount (List.length authorities)

   method set_authorities a =
       authorities <- a

   method add_additionals oad =
       additionals <- additionals@oad;
       header#set_arcount (List.length additionals)

   method set_additionals a =
       additionals <- a

   (** Output the dns packet's content as a string. *)
   method output =
      let output_all = fun str obj -> str ^ obj#output in
      header#output^
      (List.fold_left output_all "" questions) ^
      (List.fold_left output_all "" responses) ^
      (List.fold_left output_all "" authorities) ^
      (List.fold_left output_all "" additionals)
end;;

(** Stream parsers than build dns objects. *)
class message_parser =
  object (self)

    val mutable table = []
    val mutable unknown_table = []

    method private add_label elt = table <- table@[elt]

    method private get_name_from_unknown offset =
      let rec search_name liste from =
        match liste with
        | (o, e, data)::l when o <= from && e >= from ->
            let length = Char.code data.[from - o] in
            if length != 0
            then
              match length lsr 6 with
              | 0 -> 
                  begin
                    (String.sub data (from - o + 1) length) ^ "."
                    ^ self#get_name_from_unknown (from + length + 1);
                  end
              | 0b11 -> (* A pointer, cf. rfc 1035, section 4.1.4. *)
                  self#get_name (256 * (length land 0x3F)
                  + Char.code (data.[from - o + 1]))
              | nbr ->
                  raise (Malformed ("A label's length octet begins with '"
                  ^ (string_of_int nbr)
                  ^ "'. This combination is reserved for future use."))
             else 
               ""
        | (_, _, _)::l -> search_name l from
        | [] ->
            raise (Malformed ("#get_name: There is no name at this \ 
            offset ("^(string_of_int from)^") !!"))
      in
      search_name unknown_table offset;

    method private get_name offset =
      let rec search_name liste from =
        match liste with
        | (o, n, e)::l when o = from && e = 0 ->
            n ^ "."
        | (o, n, e)::l when o = from ->
            n ^ "." ^ search_name table e
        | (_, _, _)::l -> search_name l from
        | [] ->
            self#get_name_from_unknown offset
      in
      search_name table offset;

      (** Parse an IPv6 address. *)
    method private parse_addr6 stm =
      (* TODO: add compression: the use of "::" indicates one or more groups of
       * 16 bits of zeros.
       * The "::" can only appear once in an address.
       *)
      let format_16bits strong weak separator =
        match strong, weak, separator with
        | 0, 0, sep ->
            "0" ^ sep
        | 0, w, sep when w < 0x10 ->
            Printf.sprintf "%.1x%s" w sep
        | 0, w, sep ->
            Printf.sprintf "%.2x%s" w sep
        | s, w, sep when s < 0x10 ->
            Printf.sprintf "%.1x%.2x%s" s w sep
        | s, w, sep ->
            Printf.sprintf "%.2x%.2x%s" s w sep
      in
      try 
        let rec aux buf i =
          begin
            let strong = Char.code (Stream.next stm) in
            let weak = Char.code (Stream.next stm) in
            match i with
            | 7 -> 
                buf ^ format_16bits strong weak ""
            | _ ->
                aux (buf ^ format_16bits strong weak ":") (i + 1)
          end in
        aux "" 0
        with
        | Stream.Failure ->
            raise (Malformed ("#parse_addr6: not enough data."))

(** Parse an IPv4 address. *)
    method private parse_addr stm =
      try
        let one = Char.code (Stream.next stm)
        and two = Char.code (Stream.next stm)
        and thr = Char.code (Stream.next stm)
        and fou = Char.code (Stream.next stm) in
        (string_of_int one) ^ "." ^(string_of_int two)^
        "." ^ (string_of_int thr) ^ "." ^ (string_of_int fou)
      with
      | Stream.Failure ->
          raise (Malformed ("#parse_addr: not enough data."))

    (** Parse a 'dns encoded' name (rfc1035, section 4.1.2). *)
    method private parse_name stm =
      try 
        let buf = Buffer.create 63
        and start = ref (Stream.count stm) 
        and length = ref (Char.code (Stream.next stm)) in
        if !length != 0
        then
          match !length lsr 6 with
          | 0 -> 
              begin
                let rec read_label remaining_length =
                  match remaining_length with
                  | 0 ->
                      let temp = String.lowercase (Buffer.contents buf) in
                      let next_length = Char.code (List.hd (Stream.npeek 1 stm)) in
                      (match next_length with
                      | l when l lsr 6 = 0b11 -> 
                          let actual_next = 256 * (l land 0x3F) + Char.code (List.nth (Stream.npeek 2 stm) 1) in
                          self#add_label (!start, temp, actual_next)
                      | 0 ->
                          self#add_label (!start, temp, 0)
                      | _ ->
                          self#add_label (!start, temp, Stream.count stm)
                          );
                      temp
                  | _ ->
                      Buffer.add_char buf (Stream.next stm);
                      read_label (remaining_length - 1)
                in 
                let label = (read_label !length) in
                label ^ "." ^ (self#parse_name stm)
              end
            | 0b11 -> (* A pointer, cf. rfc 1035, section 4.1.4. *)
                self#get_name (256 * (!length land 0x3F) + Char.code (Stream.next stm))
            | nbr ->
                raise (Malformed ("A label's length octet begins with '"
                ^ (string_of_int nbr)
                ^ "'. This combination is reserved for future use."))
        else 
          ""
        with
        | Stream.Failure ->
            raise (Malformed ("#parse_name: not enough data."))
        | Invalid_argument "List.nth" ->
            (* XXX: Stream.npeek won't raise any error but only returns less
             * data than expected if there is not enough.
             * Hence I have to take care of error in the return List. *)
            raise (Malformed ("#parse_name: not enough data."))
        | Failure "hd" ->
            raise (Malformed ("#parse_name: not enough data."))

    (** Parse a character string (rfc1035, section 3.3). *)
    method private parse_char_string stm =
      try
        let size = Char.code (Stream.next stm) in
        let str = String.create size in
        for i = 0 to size - 1
        do
          str.[i] <- Stream.next stm;
        done;
        (*String.lowercase*) str
      with
      | Stream.Failure -> raise (Malformed ("#parse_char_string: \
      not enough data."))

      (** Parse anything, up to 65535 octets maximum. *)
    method private parse_anything size stm =
      try
        let str = String.create size in
        for i = 0 to size - 1
        do
          str.[i] <- Stream.next stm;
        done;
        String.lowercase str (* return a char_string *)
      with
      | Stream.Failure -> raise (Malformed ("#parse_char_string: \
      not enough data."))

    (** Parse a dns header. *)
    method private parse_header stm header =
      let lstf = [header#set_id; header#set_flags; header#set_qdcount; header#set_ancount;
      header#set_nscount; header#set_arcount] in
      try
        for i = 0 to 5
        do
          let fst = Char.code (Stream.next stm) 
          and snd = Char.code (Stream.next stm) in
          (List.nth lstf i) (fst * 256 + snd);
        done;
      with
      | Stream.Failure ->
          raise (Malformed ("#parse_header: not enough data ("
          ^ (string_of_int (Stream.count stm)) ^ " bytes)."))

      (** Parse a dns question. *)
    method private parse_query stm = 
      let q = new question in
      q#set_qname (self#parse_name stm);
      let lstf = [ q#set_qtype; q#set_qclass ] in
      try
        for i = 0 to 1
        do
          let fst = Char.code (Stream.next stm) 
          and snd = Char.code (Stream.next stm) in
          (List.nth lstf i) (fst * 256 + snd)
        done;
        q
      with
      | Stream.Failure ->
          raise (Malformed ("#parse_query: not enough data ("
          ^ (string_of_int (Stream.count stm)) ^ " bytes)."))

      (** Parse a dns response. *)
    method private parse_response stm = 
      let r = new rr in
      try
        r#set_name (self#parse_name stm);
        (* type *)
        let fst = Char.code (Stream.next stm) 
        and snd = Char.code (Stream.next stm) in
        let expected_type = fst * 256 + snd in
        (* class *)
        let fst = Char.code (Stream.next stm) 
        and snd = Char.code (Stream.next stm) in
        r#set_class (fst * 256 + snd);
        (* ttl *)
        let fst = Char.code (Stream.next stm) 
        and snd = Char.code (Stream.next stm)
        and thr = Char.code (Stream.next stm) 
        and qua = Char.code (Stream.next stm) in
        r#set_ttl (fst * 256 * 256 * 256 + snd * 256 * 256 + thr * 256 + qua);
        (* rdlength *)
        let fst = Char.code (Stream.next stm) 
        and snd = Char.code (Stream.next stm) in
        let expected_rdlength = fst * 256 + snd in
        (* rdata *)
        let before_rdata_offset = Stream.count stm in
        (* TODO: parse_rdata expected_length *)
        (match expected_type with
        | 1 (* A *) ->
            r#set_rdata (A (self#parse_addr stm)) (* OK *)
        | 2 (* NS *) ->
            r#set_rdata (NS (self#parse_name stm)) (* OK *)
        | 5 (* CNAME *) ->
            r#set_rdata (CNAME (self#parse_name stm)) (* OK *)
        | 6 (* SOA *) -> 
            (let get_32 stme = 
              let nstme = Tools.extract_stream stme 4 in 
              let fst = Char.code (Stream.next nstme) 
              and snd = Char.code (Stream.next nstme)
              and thr = Char.code (Stream.next nstme)
              and fou = Char.code (Stream.next nstme)
              in (fst*256+snd)*256*256+thr*256+fou
              in
              let mname = self#parse_name stm in
              let rname = self#parse_name stm in
              let serial = get_32 stm in
              let refresh = get_32 stm in
              let retry = get_32 stm in
              let expire = get_32 stm in
              let min = get_32 stm in
              r#set_rdata (SOA (mname, rname, serial, refresh, retry, expire, min)))
        | 10 (* NULL *) -> (* OK *)
            r#set_rdata (NULL (self#parse_anything expected_rdlength stm))
        | 11 (* WKS *) -> (* OK *)
            let addr = self#parse_addr stm in
            let protocol = Char.code (Stream.next stm) in
            let bitmap = self#parse_anything (expected_rdlength - 5) stm in
            let rec get_ports bit ports =
                match bit with
                | -1 -> ports
                | i ->
                    if (Char.code bitmap.[i/8]) lsr (7 - (i mod 8)) land 0x01 = 1
                    then
                        get_ports (bit - 1) (i::ports)
                    else
                        get_ports (bit - 1) ports
            in
            let ports = get_ports ((String.length bitmap) * 8 - 1) [] in
            r#set_rdata (WKS (addr, protocol, ports))
        | 12 (* PTR *) -> (* OK *)
            r#set_rdata (PTR (self#parse_name stm))
        | 13 (* HINFO *) -> (* OK *)
            let cpu = self#parse_char_string stm
            and os = self#parse_char_string stm in
            r#set_rdata (HINFO (cpu, os)) 
        | 15 (* MX *)  -> (* OK *)
            let fst = Char.code (Stream.next stm) 
            and snd = Char.code (Stream.next stm) in
            r#set_rdata (MX ((fst * 256 + snd), (self#parse_name stm)));
            (* ) *)
        | 16 (* TXT *) -> (* OK *)
            let rec get_all_char_strings cstrings =
              match Stream.count stm - before_rdata_offset with
              | current_count when current_count < expected_rdlength ->
                  get_all_char_strings ((self#parse_char_string stm)::cstrings)
              | _ ->
                  List.rev cstrings
            in
            r#set_rdata (TXT (get_all_char_strings []))
        | 28 (* AAAA *) -> (* OK *)
            r#set_rdata (AAAA (self#parse_addr6 stm))
        | 33 (* SRV *) -> (* OK *)
            let get_16 stme = 
              let nstme = Tools.extract_stream stme 2 in 
              let fst = Char.code (Stream.next nstme) 
              and snd = Char.code (Stream.next nstme)
              in fst * 256 + snd in
            let prio = get_16 stm in
            let weight = get_16 stm in
            let port = get_16 stm in
            let target = self#parse_name stm in
            r#set_rdata (SRV (prio, weight, port, target))
        | _ -> (* OK *)
            let raw_rdata = String.create expected_rdlength in
            let rec fill_rdata = function
              | over when over >= expected_rdlength -> raw_rdata
              | index ->
                  raw_rdata.[index] <- Stream.next stm;
                  fill_rdata (index + 1)
            in
            r#set_rdata (Unsupported_RDATA (expected_type, fill_rdata 0));
            unknown_table <-
              (before_rdata_offset, Stream.count stm, raw_rdata)::unknown_table;
        );
        if (Stream.count stm) - before_rdata_offset != expected_rdlength
        then
          raise (Malformed ("#parse_response: announce rdlength ("
          ^ (string_of_int expected_rdlength)
          ^ " bytes) not conform to actual size of rdata ("
          ^ (string_of_int ((Stream.count stm) - before_rdata_offset))
          ^ " bytes)."));
        r
        with
        | Tools.Toosmall _ | Stream.Failure ->
            raise (Malformed ("#parse_response: not enough data."))

            (** Parse a dns packet. *)
    method private parse_packet stm =
      let answer = new message in
      self#parse_header stm (answer#header);
      let hd = answer#header in
      (* parsing query *)
      for i = 1 to hd#qdcount
      do
        answer#add_questions [self#parse_query stm];
      done;
      (* parsing response *)
      for i = 1 to hd#ancount
      do
        answer#add_answers [self#parse_response stm];
      done;
      (* parsing auth. nameservers *)
      for i = 1 to hd#nscount
      do
        answer#add_authorities [self#parse_response stm];
      done;
      (* parsing add. info *)
      for i = 1 to hd#arcount
      do
        answer#add_additionals [self#parse_response stm];
      done;
      table <- [];
      (* TODO: if any data is remaining in the stream, and the response is
       * truncated, then I begin a new packet parsing.
       * Else then there is a parse error and I must raise an error.
       *)
      answer

end;;

(*******************: CACHING ************************
 * See in particular section 7.1 and 7.4 of RFC 1035. *********)
module Cache = Map.Make (String_triplet);;
let rr_cache = ref Cache.empty;;
let cache_lock = Mutex.create ();;

 (* Cache saving:
     * - SOA must never be saved (they normally will always have 0 TTL, but if
         * they haven't, this is an error, and I don't save anyway).
     * - If one of the rdata has a null ttl, I must discard ALL rdata OF THE
     * SAME TYPE. For instance it can be volatile information, so next time I
     * query the same way, I will want to get the right updated data. But other
     * type's data, even from a same query might be useful for another kind of
     * query. For instance a '*' query will typically return SOA data, that I
     * won't cache, but it does not mean the returned A data cannot be used for
     * a A query.
     * -> actually, from the rfc:
         * Note that using the timestamp is superior to using a current
         * time, since it allows RRs with TTLs of zero to be entered in
         * the cache in the usual manner, but still used by the
         * current
         * request, even after intervals of many seconds due to
         * system
         * load, query retransmission timeouts, etc.
     * - Finally I save rdatas with an absolute TTL. The maping key will be the
     * target, because this is how we will look for data. If we were saving by
     * type for instance, then afterwards we might have to process a long list
     * of other 
     * Note that target may mean different things for different rdata. For
     * instance in SRV, the target is got from removing service and protocol for
     * the rdata name. This is a "zone" notion. *)
 (* TODO: take query into account. Sometimes I must not save, for instance A in
* a reverse query!
* Also sometimes I must remove rrs on queries which returned no answer. For
* instance a A query which returns an empty list (not an error) implies that all
* A in cache for this name should be erased. *)
let cache_save qtime qname qtype qclass rrs =
    (* I gather the records by (name, type, class). *)
    let rec gather sep = function
        | [] -> sep
        | rr::l ->
                let like, diff =
                    List.partition (fun rr2 ->
                        rr2#name = rr#name && rr2#_type = rr#_type
                        && rr2#_class = rr#_class) l
                in gather ((rr::like)::sep) diff
    in let gathered_rrs = gather [] rrs
    in let rec absolute_ttl_rrs storable = function
        | [] -> storable
        | rr::l ->
            (* TODO: actually check that a TTL is no more than a week and either
             * discard the whole response or cap the TTL. Cf.
             * section 7.3 fo RFC 1035*)
            absolute_ttl_rrs (((float_of_int rr#ttl) +. qtime, rr)::storable) l
    in let clean qt =
        (* There is normally no need to "clean" as this will be done when
         * updating the cache... except the case when the answer section was
         * empty! But this is still update information for the cache. Indeed if
         * there was some records for a type but they have been erased, I
         * need to update my cache accordingly. This function do this explicit
         * cleaning. *)
        Mutex.lock cache_lock;
        rr_cache := Cache.remove (qname, qt, qclass) !rr_cache;
        Mutex.unlock cache_lock
    in let clean_any () =
        Mutex.lock cache_lock;
        rr_cache := Cache.filter
            (fun (qn, qt, qc) rr -> qname != qn || qclass != qc)
            !rr_cache;
        Mutex.unlock cache_lock
    in let rec store = function
        | [] -> ()
        | []::_ ->
            (* Is supposed to be impossible. *)
            failwith "Unexpected error in cache_save\n";
        | (rr::rrs)::l ->
            (
            match rr#rdata with
            | (Unsupported_RDATA _) -> ()
            | (SOA _) ->
                (* I save the SOA records, but I override any ttl by 0. In
                 * fact, this is equivalent of not saving but reusing the
                 * same SOA if a query asked before is actually processed
                 * after. *)
                Mutex.lock cache_lock;
                rr_cache :=
                    Cache.add (rr#name, rr#_type, rr#_class) [(qtime, rr)] !rr_cache;
                    (* Normally if there are 2 SOA, they are the same, for
                     * instance in an axfr query. *)
                Mutex.unlock cache_lock;
            | _ ->
                Mutex.lock cache_lock;
                rr_cache :=
                    Cache.add (rr#name, rr#_type, rr#_class)
                    (absolute_ttl_rrs [] (rr::rrs)) !rr_cache;
                Mutex.unlock cache_lock;
            );
            store l
    in
    (* Now we see what has to be cached and how. *)
    match qname, qtype, qclass with
    (* Two vague queries are no good for cache. *)
    | qn, _, _ when String.contains qn '*' -> ()
    | _, _, 255
    (* I don't cache obsolete records. Two few possibilities to ever be of any
     * use. *)
    | _, 10, _ (* NULL *)
    | _, 11, _ (* WKS, deprecated in rfc 1127 *)
    | _, 3, _ (* MD *)
    | _, 4, _ (* MF *) -> ()
    | _, qt, _ when (qt > 0 && qt < 16) || qt = 33 || qt = 28 ->
            (* Single type queries. *)
            clean qt;
            store gathered_rrs
    | _, 253, _ (* MAILB = MB + MG + MR *) ->
            clean 7;
            clean 8;
            clean 9;
            store gathered_rrs
    | _, 255, _ (* ANY, '*' *) ->
            clean_any ();
            store gathered_rrs
    | _, _, _ -> ()
;;

(* Cache query implementation:
    * - If SOA is asked, return empty list.
    * - If type/class unknown, I return empty list.
    * - for one type of query, I look for any data.
    * - If I find some, and all have good ttl, I consider having my data. I
    * return them, but transforming the ttls into the relative remaining time
    * before expiration.
    * - If one of the rdata has a perempted TTL, I discard ALL rdata for the
    * given query and return empty list (hence in a same query, the smaller tll
    * takes precedence over others). This is because if I were to discard only
    * this rdata, it would not be returned to the user, which is not good,
    * because the fact that the record expired does not mean the resource does
    * not exist anymore. This must be checked against the source anyway, so
    * let's discard all and get new data to cache.
    *)
(* Return the triplet of answer, authority and additional records. *)
let cache_request qtime qname qtype qclass =
    let record_search qn qt qc =
        try
            let zone_rrs = Cache.find (qn, qt, qc) !rr_cache
            in
            let rec check_ttl result = function
                | [] -> result
                | (time, rr)::l when time < qtime ->
                        (* One of the records is expired. I must not return a
                         * partial list because it is highely likely that the record
                         * is renewed. I must query the nameservers and discard all
                         * data of this name/type/class.
                         *)
                        Mutex.lock cache_lock;
                        rr_cache := Cache.remove (qname, qt, qc) !rr_cache;
                        Mutex.unlock cache_lock;
                        []
                | (time, rr)::l ->
                        rr#set_ttl (int_of_float (time -. qtime));
                        check_ttl (rr::result) l
            in 
            check_ttl [] zone_rrs
        with
        | Not_found | Unknown_type _ | Unknown_class _ ->
            (* Records of unknown class or types are not saved *)
            []
    in
    let sub_request qn qt qc =
        match qt with
        | 7 (* MB *) ->
                (* MB records cause additional section processing which looks up
                 * an A type RRs corresponding to MADNAME.
                 * And AAAA as extended by RFC 3596 *)
                let answers = record_search qn qt qc in
                let rec aux current_adds = function
                    | [] -> current_adds
                    | rr::l ->
                            (* It is assured by record_search that I will have
                             * only records of type MB. 
                             * I can then safely use a polymorphic variant. *)
                            let addr =
                                match rr#rdata with
                                | MB a -> a
                                | _ -> 
                                        failwith "function sub_request: \
                                        impossible exception for MB.\n"
                            in
                            let addr_a = record_search addr 1 qc in
                            let addr_aaaa = record_search addr 28 qc in
                            let addrs = addr_a@addr_aaaa in
                            (* As personal implementation choice, if the cache
                             * is not enough to fill the additionals, I bypass
                             * the cache. *)
                            if addrs = []
                            then
                                failwith "ADDR"
                            else
                                aux (addrs@current_adds) l
                in
                begin
                    try
                        answers, [], aux [] answers
                    with
                    | Failure "ADDR" -> [], [], []
                end
        | 15 (* MX *) ->
                (* MX records cause type A additional section processing for the
                 * host
                 * specified by EXCHANGE *)
                let answers = record_search qn qt qc in
                let rec aux current_adds = function
                    | [] -> current_adds
                    | rr::l ->
                            let addr =
                                match rr#rdata with
                                | MX (_, a) -> a
                                | _ -> 
                                        failwith "function sub_request: \
                                        impossible exception for MX.\n"
                            in
                            let addr_a = record_search addr 1 qc in
                            let addr_aaaa = record_search addr 28 qc in
                            let addrs = addr_a@addr_aaaa in
                            (* As personal implementation choice, if the cache
                             * is not enough to fill the additionals, I bypass
                             * the cache. The NS may have more information. *)
                            if addrs = []
                            then
                                failwith "ADDR"
                            else
                                aux (addrs@current_adds) l
                in
                begin
                    try
                        answers, [], aux [] answers
                    with
                    | Failure "ADDR" -> [], [], []
                end
        | 33 (* SRV *) ->
                (* Implementors are urged, but not required,
                 * to return the address record(s) in the
                 * Additional Data section. *)
                let answers = record_search qn qt qc in
                let rec aux current_adds = function
                    | [] -> current_adds
                    | rr::l ->
                            let addr =
                                match rr#rdata with
                                | SRV(_, _, _, a) -> a
                                | _ -> 
                                        failwith "function sub_request: \
                                        impossible exception for SRV.\n"
                            in
                            let addr_a = record_search addr 1 qc in
                            let addr_aaaa = record_search addr 28 qc in
                            let addrs = addr_a@addr_aaaa in
                            (* As personal implementation choice, if the cache
                             * is not enough to fill the additionals, I bypass
                             * the cache. The NS may have more information. *)
                            if addrs = []
                            then
                                failwith "ADDR"
                            else
                                aux (addrs@current_adds) l
                in
                begin
                    try
                        answers, [], aux [] answers
                    with
                    | Failure "ADDR" -> [], [], []
                end
        | 2 (* NS *) ->
                (* NS records cause both the usual additional section processing
                 * to locate
                 * a type A record, and, when used in a referral, a special
                 * search of the
                 * zone in which they reside for glue information.
                 * -> XXX: I am not sure to understand the second point.
                 * TODO?
                 *)
                let answers = record_search qn qt qc in
                let rec aux current_adds = function
                    | [] -> current_adds
                    | rr::l ->
                            let addr =
                                match rr#rdata with
                                | NS a -> a
                                | _ -> 
                                        failwith "function sub_request: \
                                        impossible exception for NS.\n"
                            in
                            let addr_a = record_search addr 1 qc in
                            let addr_aaaa = record_search addr 28 qc in
                            let addrs = addr_a@addr_aaaa in
                            (* As personal implementation choice, if the cache
                             * is not enough to fill the additionals, I bypass
                             * the cache. The NS may have more information. *)
                            if addrs = []
                            then
                                failwith "ADDR"
                            else
                                aux (addrs@current_adds) l
                in
                begin
                    try
                        answers, [], aux [] answers
                    with
                    | Failure "ADDR" -> [], [], []
                end
        | _ ->
                (* Other type of supported records have no additional record
                 * processing. *)
                record_search qn qt qc, [], []
    in
    (* Now we see what has to be queried. *)
    match qname, qtype, qclass with
    (* ANY kind of wildcard query must be actually queried as any presence in
    * cache of some record does not prove (in)existence of other data. *)
    | qn, _, _ when String.contains qn '*' ->
            [], [], []
    | _, 255, _(* type '*' *)
    | _, _, 255 (* class '*' *)
    (* I don't cache obsolete data as well. *)
    | _, 3, _ (* MD *)
    | _, 4, _ (* MF *)
    | _, 11, _ (* WKS, deprecated in rfc 1127 *)
    | _, 10, _ (* NULL *) ->
            [], [], []
    | _, qt, _ when (qt > 0 && qt < 16) || qt = 33 || qt = 28 ->
            (* Single type queries. *)
            sub_request qname qtype qclass
    | _, 253, _ (* MAILB = MB + MG + MR *) ->
            let mb, _, mbadd = sub_request qname 7 qclass in
            let mg, _, _ = sub_request qname 8 qclass in
            let mr, _, _ = sub_request qname 9 qclass in
            (* Will return only if ALL 3 type of records return data.
             * Even if any is empty in the cache,
             * it may have data on the name server. *)
            if mb != [] && mg != [] && mr != []
            then
                mb@mg@mr, [], mbadd
            else
                [], [], []
    | _, _, _ ->
            [], [], []
;;

(************************ RESOLVERS ************************************)

(** Class that permits a service abstraction *)
class virtual dns_service =
object (self)
   inherit message_parser

   val mutable nameservers = []
   val mutable search = []
   val port = 53
   val mutable timeout = 5.
   val mutable retries = 1
   val mutable verbose_output = None
   val mutable ansi_capable = false

   val mutable query = new message 
   val mutable answer = None

(**************** VERBOSE METHODS ******************)
   method verbose = verbose_output
   method set_verbose (output: out_channel option) =
       verbose_output <- output

   method ansi_capable = ansi_capable

   method set_ansi_capable value =
        ansi_capable <- value

   (* TODO: method print_verbose: ('a, out_channel, unit) format -> 'a
    * printf format! *)
   method private print_verbose fmt =
       (* TODO: add color when output = stdout or stderr? *)
       try match verbose_output with
       | None -> ()
       | Some oc ->
               (*ignore (Printf.fprintf oc (format_of_string fmt));
               ()*)
               output_string oc fmt
       with
       |_ -> ()
       (* I don't want the debug feature being able to crash the program. That
        * would be pretty ironic! *)

    method private print_retry i =
        if (ansi_capable)
        then
            begin
                (* XXX: I don't know how to insert non printable
                 * characters in OCaml's strings. Is it possible
                 * directly? *)
                (* I set retry text in red. *)
                let esc = "^[0;31m" in
                esc.[0] <- Char.chr 27;
                self#print_verbose esc;
          end;
        self#print_verbose (Printf.sprintf "\n== Retry %d ==\n" i);
        if (ansi_capable)
        then
            begin
                let esc = "^[0m" in
                esc.[0] <- Char.chr 27;
                self#print_verbose esc;
          end;

    method private print_query_begin ns =
        if (ansi_capable)
        then
            begin
                let esc = "^[0;1m" in
                esc.[0] <- Char.chr 27;
                self#print_verbose esc;
            end;
        self#print_verbose (Printf.sprintf "* Trying name server \
        %s\n" ns);
        if (ansi_capable)
        then
            begin
                let esc = "^[0m" in
                esc.[0] <- Char.chr 27;
                self#print_verbose esc;
            end;

    method private print_error text =
        if (ansi_capable)
        then
            begin
                (* XXX: I don't know how to insert non printable
                 * characters in OCaml's strings. Is it possible
                 * directly? *)
                (* I set error text in magenta. *)
                let esc = "^[0;35m" in
                esc.[0] <- Char.chr 27;
                self#print_verbose esc;
          end;
        self#print_verbose text;
        if (ansi_capable)
        then
            begin
                let esc = "^[0m" in
                esc.[0] <- Char.chr 27;
                self#print_verbose esc;
          end;
(**************** END OF VERBOSE METHODS ******************)

   method nameservers = nameservers
   method set_nameservers = function
       | [] -> 
               let servers, _, _ = Tools.resolv_conf () in
               nameservers <- servers
       | ns -> nameservers <- ns

   method search_list = search
   method set_search_list = function
       | [] ->
               let _, _, servers = Tools.resolv_conf () in
               search <- servers
       | l -> search <- l

   method raw_query = query
   method raw_response = answer

   method timeout = timeout
   method set_timeout new_timeout = timeout <- new_timeout

   method retries = retries
   method set_retries n = retries <- max 0 n

    method recursive =
        query#header#is_rd

    method set_recursive value =
        query#header#set_rd value

   (* Send the query packet.
    * @param sck is a file descriptor (file_descr). *)
   method private send sck = 
      let oc = Unix.out_channel_of_descr sck in
      output_string oc query#output;
      flush oc

   (** Send the answer packet.
    * @param sck is a file descriptor (file_descr). *)
   method private receive sck =
      let str = String.create 512 in
      (* XXX: Messages carried by UDP are restricted to 512 bytes (not counting
       * the IP or UDP headers). *)
      let size, void = Unix.recvfrom sck str 0 512 [] in
      let stm = Stream.of_string (String.sub str 0 size) in
      answer <- Some (self#parse_packet stm)

   (** Method that will do the work *)   
   method virtual query : unit

   initializer
       let servers, _, sear = Tools.resolv_conf () in
       nameservers <- servers;
       search <- sear
end
;;

exception Name_error;;
exception Nameservers_failure;;
exception Not_queried;;

(** Query class like in perl Net::Dns *)
class resolver =
    object (self)
        inherit dns_service as super

        (* Is false when "query" has not been run or the query changed since
         * the last query. *)
        val mutable answered = false
        (* failure is true if all nameservers failed OR timed-out. *)
        val mutable failure = false
        (* name_error is true if an authoritative name server told me this
         * domain name does not exist. *)
        val mutable name_error = false
        val mutable nocache = false
        val mutable cached_answers = []
        val mutable cached_additionals = []
        val mutable cached_authorities = []

        method nocache = nocache 
        method set_nocache nc = nocache <- nc

        method private set_unanswered =
            cached_answers <- [];
            cached_additionals <- [];
            cached_authorities <- [];
            answer <- None;
            answered <- false;
            name_error <- false;
            failure <- false

        method private answered =
            match answered, failure, name_error with
            | false, _, _ -> raise Not_queried
            | true, true, _ -> raise Nameservers_failure
            | true, _, true -> raise Name_error
            | _ -> ()

        method reset_queries =
            query#set_questions [];
            self#set_unanswered

        (** Add a query to the packet with name type and class. *)
        method add_query ?(qtype: string = "A")
        ?(qclass: string = "IN")
        ~(qname: string) 
        =
            let ret = new question in
            ret#set_qname qname;
            ret#set_qtype (string_to_qtype qtype);
            ret#set_qclass (string_to_qclass qclass); 
            query#add_questions [ret];
            self#set_unanswered

        (** Query the nameserver. *)
        method query =
            (* When I make a new query, I reset any previous answer. *)
            self#set_unanswered;
            (* I select the domain PF_INET6 which should be compatible with both
             * IPv4 and IPv6. *)
            let rec lookup retry = function
            | [] ->
                    if retry > 0
                    then
                        (
                            self#print_retry (retries - retry + 1);
                            lookup (retry - 1) nameservers 
                        )
                    else
                        (* All name servers failed or timed-out. *)
                        (
                            self#print_error "\nAll name servers failed or \
                            timed-out\n";
                            answered <- true;
                            failure <- true
                        )
            | s::l ->
                    self#print_query_begin s;
                    let sck = Unix.socket Unix.PF_INET6 Unix.SOCK_DGRAM 0 in
                    Unix.connect sck
                        (Unix.ADDR_INET (Unix.inet_addr_of_string s, port));
                    super#send sck;
                    try
                        match Unix.select [sck] [] [] timeout with
                        | [], _, _ ->
                                raise TimedOut
                        | fd::_, _ , _ -> 
                                super#receive fd;
                                match self#raw_response with
                                | None ->
                                        (* The response was invalid. *)
                                        self#print_error "\tInvalid response \
                                        received. Probably a format issue.\n";
                                        lookup retry l
                                | Some a ->
                                        (* If I have no answer and an error code
                                         * and this is not an authority answer
                                         * telling me the name does not exist,
                                         * then I loop to other nameservers. *)
                                        if a#header#rcode != 0
                                        then begin
                                            if a#header#rcode = 3
                                            && a#header#is_aa
                                            then
                                                (
                                                answered <- true;
                                                name_error <- true;
                                                )
                                            else
                                                (
                                                    (*answer <- None;*)
                                                self#print_error
                                                (Tools.rcode_to_string a#header#rcode);
                                                lookup retry l
                                                )
                                        end
                                        else
                                            answered <- true
                                        (* we got our answer,
                                         * it is finished. *)
                    with
                    | TimedOut ->
                            self#print_error "\tTimed-Out\n";
                            lookup retry l
            in
            let qtime = Unix.time () in
            (* I only take into account single question queries in the cache. *)
            let caching = List.length query#questions = 1 && not nocache in
            if caching
            then
                begin
                    let question = List.hd query#questions in
                    let answers, additionals, authorities =
                        cache_request qtime question#qname
                        question#qtype question#qclass in
                    if List.length answers != 0
                    then
                        begin
                            answered <- true;
                            cached_answers <- answers;
                            cached_additionals <- additionals;
                            cached_authorities <- authorities
                        end
                end;
            if not answered
            then
                begin
                    lookup retries nameservers;
                    (* TODO: @ add @ auth *)
                    if caching
                    then
                        try
                            let question = List.hd query#questions in
                            cache_save qtime question#qname question#qtype
                            question#qclass self#answers
                        with
                        (* If a problem occurred during query, there is nothing
                         * which can be deducted for cache update. *)
                        | _ -> ()
                end

        method answers =
            self#answered;
            match (cached_answers, super#raw_response) with
            | c, _ when List.length c > 0 ->
                    self#print_verbose "Cached Data:\n";
                    c
            | _, None -> []
            | _, Some a ->
                    a#answers

        method authorities =
            self#answered;
            match cached_authorities, super#raw_response with
            | c, _ when List.length c > 0 -> c
            | _, None -> []
            | _, Some a ->
                   a#authorities

        method additionals =
            self#answered;
            match cached_additionals, super#raw_response with
            | c, _ when List.length c > 0 -> c
            | _, None -> []
            | _, Some a ->
                    a#additionals

    end;;

let resolver_from_model model =
    let r = new resolver in
    r#set_nameservers (model#nameservers);
    r#set_timeout (model#timeout);
    r#set_retries (model#retries);
    r#set_recursive (model#recursive);
    r#set_search_list (model#search_list);
    r#set_verbose (model#verbose);
    r#set_ansi_capable (model#ansi_capable);
    r
;;

(** Do axfr request. *)
class dns_axfr =
object (self)
  inherit dns_service as super

  val mutable soa = 0
  val mutable leftstream = ""
  val mutable lst = []

  method get_lst = lst

   (** List this domain. *)
  (* TODO: why is there no class here? As far as I can see, even AXFR queries
   * can have classes other than IN. *)
   method ls (q: string) =
      let ret = new question in
      ret#set_qname (String.lowercase q);
      ret#set_qtype 252;
      ret#set_qclass 1;
      query#header#set_rd true;
      query#add_questions [ret]

   method private parse_axfr stm =
      let answer = self#parse_packet stm in
      if answer#header#ancount > 0 && (List.hd answer#answers)#_type = 6
      then
         soa <- soa + 1;
      answer

   (** Do the work. *)
   method query =
      let sck = Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0 in
      Unix.connect sck
      (Unix.ADDR_INET (Unix.inet_addr_of_string (List.hd nameservers), port));
      let oc = Unix.out_channel_of_descr sck in
      output_string oc (Tools.hexa_of_int (String.length super#raw_query#output)
      2);  (* let's send the packet size ... *)
      flush oc;
      super#send sck;
      let loop = ref true in
      while !loop
      do
         (match Unix.select [sck] ([]: Unix.file_descr list) ([]: Unix.file_descr list) timeout with
         | read, _ , _ when read != []-> 
            (let str = String.create 4096 in
            let size = Unix.handle_unix_error (Unix.recv sck str 0 4096)
            [] in
            self#break_stream (String.sub str 0 size);
            (* First elt must be a SOA ... *)
            let value = if lst = [] then false else (List.hd lst)#header#rcode = 6 in
            if soa = 2 || value 
            then
               loop := false;
            )
         | _, _, _ -> 
            (Unix.shutdown sck Unix.SHUTDOWN_ALL;
            loop := false))
      done;

   method private get_domain str =
      try
      let _ = Str.search_forward (Str.regexp "\\.\\(.*\\)$") str 0 in 
      Str.matched_group 1 str;
      with
      | Not_found -> ""
      
   method private break_stream str =
      leftstream <- leftstream^str;
      let test = ref true 
      and size = ref ((String.length leftstream))in
      while !test
      do
         try
         if !size <= 3
         then
            raise Stream.Failure
         else
            begin
            let stm = Stream.of_string leftstream in 
            let one = Char.code (Stream.next stm)
            and two = Char.code (Stream.next stm) in
            let sz = one*256+two in
            (if !size-2 >= sz 
            then
               begin
               let nstm = Tools.extract_stream stm sz in
               size := !size - sz - 2;
               leftstream <- Tools.stream_to_string stm;
               try
	       let str = Tools.stream_to_string nstm in
	       let tmp1 = Stream.of_string str in
               let pkt = self#parse_axfr tmp1 in
               lst <- lst@[pkt];
	       with
               | Tools.Toosmall inte -> test := false
	       | Unknown_type str -> print_string ("Catch: "^str^"\n");
               if soa = 2
               then
                  raise Stream.Failure
               end
            else
               raise Stream.Failure)
            end;
         with 
         | Stream.Failure -> test := false;
      done
      
end;;
