(**
 dns.mli --- Ocaml DNS library
                                                                         
 This file is part of the ODNS library.

 Copyright (C)  2010-2011  Jehan Hysseo <hysseo at zemarmot point net>
                                                                         
 This program is free software; you can redistribute it and/or modify it 
 under the terms of the GNU General Public License version 2 as          
 published by the Free Software Foundation.                              
                                                                         
 This program is distributed in the hope that it will be useful, but 
 WITHOUT ANY WARRANTY; without even the implied warranty of            
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     
 General Public License for more details.                                

 @see <http://tools.ietf.org/html/rfc1035> Domain names - Implementation and specification
 @see <http://tools.ietf.org/html/rfc2782> DNS SRV
 @see <http://www.ietf.org/rfc/rfc3596.txt> DNS Extensions to Support IP Version 6
 @author Jehan Hysseo (current author and maintainer, after version 0.1)
 @author Guillaume Valadon (original, version 0.1)
 @author Manu Letellier (original, version 0.1)
 @version 0.2

*)

(* TODO: should be caught inside the library! *)
(*exception Malformed of string;;*)
(* Will be raised if a packet is malformed (for example, too small). *)

(*val string_to_qtype: string -> int;;*)
(* [string_to_qtype type] transforms a type in usual litteral form into the
 * corresponding integer DNS question type.
 * @param type is the type in litteral form.
 * @return the integer value used in DNS queries.
 * *)

(*********************************)
(** {1 Low Level DNS elements} *)
(*********************************)

(** Generic classes to easily construct raw DNS message from scratch. 
 Some knowledge of the DNS format is required here. You might probably want to look the
 last section for higher level DNS query constructors. *)

(** {2 Sections of a DNS Message} *)

(** {3 Header} *)

(** QR field of message header. *)
type message_type =
    | Query
    | Response;;

(** Represents a DNS message's header. *)
class header :
object 
   method id : int
   (** 16 bit message identifier *)
   method set_id : int -> unit 
   (** Manually set the identifier (an id is randomly generated by default
    otherwise). *)

   method qr : message_type
   (** QR bit: Question or Response? *)
   method set_qr : message_type -> unit
   (** Manually set the QR bit (default to [Question]). *)

   method opcode : int
   (** @return the kind of query (Operation Code on 4 bits). *)
   method set_opcode : int -> unit
   (** Manually set the kind of query (default to [0], standard query). *)

   method is_aa : bool
   (** @return [true] if the Answer is Authoritative. *)   
   method set_aa : bool -> unit
   (** Manually define an answer being authoritative (default to [false]). *)

   method is_tc : bool
   (** @return [true] if the message is TrunCated. *)
   method set_tc : bool -> unit
   (** Manually define a truncated message (default to [false]). *)

   method is_rd : bool
   (** @return [true] is Recursion is Desired by a query (copied in response). *)
   method set_rd : bool -> unit
   (** Manually set the desire of recursion (default to [true]). *)

   method is_ra  : bool
   (** @return [true] if Recursion is Available by the name server (set in
    response). *)
   method set_ra : bool -> unit
   (** Manually define the availability of recursion (default to [false]).
    
     {i Note: the current resolver is not able of recursion in its current state
     but queried name servers usually are. So setting [rd] to [true] is enough
     in most cases.}. *)

   method z: int
   (** @return the {b z} field (3 bits), which is reserved for future use (must
    be zero in all queries). *)
   method set_z: int -> unit
   (** Manually set the {b z} field.
    
    {b Warning}: use this method only when you
    know what to do. It should normally always be [0]. *)

   method rcode : int
   (** @return the Response code (4 bits, set in response). *)  
   method set_rcode : int -> unit
   (** Manually set the response code (default to [0], no error condition). *)
  
   method qdcount : int
   (** @return QDCOUNT (16 bits integer): number of entries in question section. *)
   method set_qdcount : int -> unit
   (** Manually set the number of entries in question section (default to [0]). *)

   method ancount : int
   (** @return ANCOUNT (16 bits integer): number of resource records in the answer section. *)
   method set_ancount : int -> unit
   (** Manually set the number of records in answer section (default to [0]). *)

   method nscount : int
   (** @return NSCOUNT (16 bits integer): number of name server resource records in the authority records section. *)
   method set_nscount : int -> unit
   (** Manually set the number of records in authority section (default to [0]). *)

   method arcount : int
   (** @return ARCOUNT (16 bits): number of resource records in the additional records section. *)
   method set_arcount : int -> unit
   (** Manually set the number of records in additional section (default to [0]). *)

   method set_flags : int -> unit
   (** You can set all the "flags" together through this single method
    (when you know the internal header structure).
    What is called flags here are the following fields (16 bits all together):
    QR, Opcode, AA, TC, RD, RA, Z (must always be 0) and RCODE. See RFC 1035 for details. *)
   method flags : int
   (** Get all the "flags" as a 16 bits integer. *)

   method output : string
   (** @return the header's content as a raw binary string. *)
end;;

(** {3 Question} *)

(** Represents a question section of a DNS message. *)
class question :
object
    method qname : string
    (** @return the queried domain name. *)
    method set_qname : string -> unit
    (** [set_qname dn] sets the domain name to query to [dn]. *)

    method qtype : int
    (** @return the type of query. *)
    method set_qtype : int -> unit
    (** [set_qtype t] sets the query type to [t]. *)

    method qclass : int
    (** @return the class of the query. *)
    method set_qclass : int -> unit
    (** [set_qclass c] sets the class of query to [c]. *)

    method output : string
    (** @return the dns question's content in its binary form. *)
end;;

(** {3 Resource Records (answer, authorities, additionals)} *)

(** Resource Record data. *)
type rdata = 
   | A of string
   (** {b Address} ({e Internet specific record}): IPv4 address *)
   | NS of string
   (** {b authoritative Name Server}: domain name *)
   | MD of string
   (** {b Mail Delivery}: domain name (Obsolete - use MX) *)
   | MF of string 
   (** {b Mail Forwarding}: domain name (Obsolete - use MX) *)
   | CNAME of string
   (** {b Canonical Name}: domain name *)
   | SOA of string * string * int * int * int * int * int
   (** {b Start Of Authority}:
    (primary name server for this zone, responsible's mailbox, serial,
    refresh, retry, expire, minimum) *)
   | MB of string
   (** {b Mailbox}: domain name (Experimental) *)
   | MG of string
   (** {b Mail Group}: domain name (Experimental) *)
   | MR of string
   (** {b Mailbox Rename}: maibox (Experimental) *)
   | NULL of string
   (** Anything. NULL records are not allowed in master files, they are used as
    placeholders in experimental extensions. *)
   | WKS of string * int * (int list)
   (** {b Well Known Services}: (address, IP protocol number, list of well known service's ports)

   {e Internet specific record} which has been
    deprecated (see RFC 1127). *)
   | PTR of string
   (** {b Pointer}: domain name *)
   | HINFO of string * string
   (** {b Host Info}: (cpu type, operating system type) *)
   | MINFO of string * string
   (** {b Mail Info}: (responsible mailbox, error mailbox) (Experimental) *)
   | MX of int * string
   (** {b Mail eXchange}: (preference, domain name) *)
   | TXT of string list
    (** {b Text}: list of descriptive texts. *)
   | SRV of int * int * int * string
   (** {b Service}: (priority, weight, port, target) *)
   | AAAA of string 
   (** {b Address} ({e Internet specific record}): IPv6 address *)
   | Unsupported_RDATA of int * string
   (** Unsupported type: (type number, raw rdata field) *)
;;

(** Represents a Resource Records section of a DNS message. *)
class rr :
object 
    method name : string
    (** @return the domain name to which this resource record pertains. *)
    method set_name : string -> unit
    (** [set_name dn] sets to [dn] the domain name to which this resource record
     pertains. *)


    method _class : int
    (** @return the class of the data in the rdata field. *)
    method set_class : int -> unit
    (** [set_class c] sets to [c] the class of the data which is to be set in the rdata
     field. *)

    method ttl : int
    (** @return the Time To Live, a 32 bit unsigned integer that specifies the
      time interval (in seconds) that the resource record may be cached before
      it should be discarded. Zero values are interpreted to mean that the RR
      can only be used for the transaction in progress, and should not be
      cached.
      (default: 0)
     *)
    method set_ttl : int -> unit
    (** [set_ttl t] sets to [t] the Time To Live. *)

    method rdata : rdata 
    (** @return the resource record's data. *)
    method set_rdata : rdata -> unit
    (** [set_rdata rd] sets to [rd] the resource record's data. *)

    method _type : int
    (** @return the type of resource.
     This type is computed automatically from the rdata being set with
     {!set_rdata}.
     *)
    method rdlength : int
    (** @return the length in octet (unsigned 16 bits integer) of the RDATA field.
      This length is computed automatically when rdata is inserted with
      {!set_rdata}. *)

    method output : string 
    (** @return the DNS resource record's content as a raw binary data. *)
end;;

(** {2 Raw DNS message} *)

(** Represents a DNS message. *)
class message :
object 
   method header : header
   (** @return the {!header} of the message. Be careful before modifying this
    header yourself. *)
   method questions : question list
   (** @return the list of questions of the {i question} section of the DNS
    message. *)
   method answers : rr list
   (** @return the list of Resource Records of the {i answers} section of the DNS
    message. *)
   method authorities : rr list
   (** @return the list of Resource Records of the {i authorities} section of the DNS
    message. *)
   method additionals : rr list
   (** @return the list of Resource Records of the {i authorities} section of the DNS
    message. *)

   method add_questions : question list -> unit
   (** [add_questions [q1; q2]] adds the two {!question}s [q1] and [q2] to the
   {e question section} (to be used in DNS messages of type [Query]). *)
   method set_questions: question list -> unit 
   (** [set_questions [q1; q2]] removes any question in the
   {e Question section} and replaces them by [q1] and [q2]
   (to be used in DNS messages of type [Query]). *)

   method add_answers : rr list -> unit
   (** [add_answers [a1; a2]] adds the two {!rr}s [a1] and [a2] to the
   {e Answer section} (to be used in DNS messages of type [Response]). *)
   method set_answers: rr list -> unit
   (** [set_answers [a1; a2]] removes any record currently in the 
   {e Answer section} and replaces them by the {!rr}s [a1] and [a2]
   (to be used in DNS messages of type [Response]). *)

   method add_authorities : rr list -> unit
   (** [add_authorities [a1; a2]] adds the two {!rr}s [a1] and [a2] to the
   {e Authority section} (to be used in DNS messages of type [Response]). *)
   method set_authorities: rr list -> unit
   (** [set_authorities [a1; a2]] removes any record currently in the 
   {e Authority section} and replaces them by the {!rr}s [a1] and [a2]
   (to be used in DNS messages of type [Response]). *)

   method add_additionals : rr list -> unit
   (** [add_additionals [a1; a2]] adds the two {!rr}s [a1] and [a2] to the
   {e Additional section} (to be used in DNS messages of type [Response]). *)
   method set_additionals: rr list -> unit
   (** [set_additionals [a1; a2]] removes any record currently in the 
   {e Additional section} and replaces them by the {!rr}s [a1] and [a2]
   (to be used in DNS messages of type [Response]). *)

   method output : string
   (** @return the actual dns message in binary form as
    it will be sent over the wire. *)
end;;

(** {1 Resolver} *)

(** Higher level DNS resolvers.
In most cases, these are the classes to use in order to resolve domain names in
the most powerful way. *)

exception Unknown_type of string;;
(** [Unknown_type t] will be raised 
if the user attempts to query with an unknown type [t].
It means either that the type does not exist or that it is not supported by
{b ODNS}.
*)

exception Unknown_class of string;;
(** [Unknown_class c] will be raised if the user attempts to query with an
unknown class [c].
It means either that the class does not exist or that it is not supported by
{b ODNS}.
*)

exception Name_error;;
(** Raised when the domain name does not exist. *)
exception Nameservers_failure;;
(** Raised when the query failed on all nameservers, either by timeout, or
 because the response was wrongly formatted, or finally because the response
 message had a return code in error (other than the "name error", which is an
 informational error indicating that the domain name does not exist, if returned
 by an authoritative name server).
 *)
exception Not_queried;;
(** Raised when you attempt to obtain resource records while the query has not
 been made yet. *)

(** DNS resolver. *)
class resolver:
object
    (** {4 High Level Methods} *)

   method reset_queries: unit
   (** Delete all queries.

   NOTE: when doing this, [answers], [authorities] and [additionals] are
   reseted as well as though the resolver was new.

   This methods allows the same resolver to be used multiple times with the
   same configuration.
   *)

   method add_query: ?qtype:string -> ?qclass:string -> qname:string -> unit
   (** [add_query ~qname ~qtype ~qclass] adds a new query to the packet.

   The exceptions {!Unknown_type} and {!Unknown_class} may occurre when a
   support of a query type or class is missing in {b ODNS}. If this was to
   happen, you can still construct the query by adding the type and the class as
   their integer value using {!raw_query}.

   NOTE: when adding a query, [answers], [authorities] and [additionals] are
   reseted as well as though the resolver was new, so that the answers are
   always consistent to the current list of query.

    @param qname the domain name queried.
    @param qtype the type of the query (default to ["IN"]).
    @param qclass the class of the query (default to ["A"]).
    @raise Unknown_type when [qtype] does not correspond to a known resource
    record.
    @raise Unknown_class when [qclass] does not correspond to a known class. *)

   method query: unit
   (** Actually query the nameservers (using the internal algorithm looping
    through the list of nameservers, waiting for [timeout] seconds and
    retrying [retries] times when failures occurred). *)

   (* TODO:
       * method search: unit
       * Same as {!query} but use the {!search_list}. *)

   method answers: rr list
   (**
   A resolver can be reused several time.
   Hence it will return only records returned as response of the last time you
   ran {!query}.
   Moreover this function has no meaning before you run {!query} (it would raise
   {Not_queried}).
   @return the list of resource records in the {b Answer} section of the DNS
   response.
   An empty list means that we had a valid response but that there is no record
   for the queried target, type and class (resolving errors are different and
   will raise {!Name_error} or {!Nameservers_failure}).
   @raise Name_error when a queried domain name does not exist.
   @raise Nameservers_failure when all nameservers failed us and we got no
   response (which does not mean there is none. We just don't know).
   @raise Not_queried when you try to get answers while you haven't actually
   queried yet (with {!query}) or whether you have changed the query since the
   last {!query} (with {!add_query} or {!reset_queries}).
   *)
   method authorities: rr list
   (**
   A resolver can be reused several time.
   Hence it will return only records returned as authorities of the last time you
   ran {!query}.
   Moreover this function has no meaning before you run {!query} (TODO: raise
   an exception then).
   @return the list of resource records in the {b Authority} section of the DNS
   response.
   @raise Name_error when a queried domain name does not exist.
   @raise Nameservers_failure when all nameservers failed us and we got no
   response (which does not mean there is none. We just don't know).
   @raise Not_queried when you try to get answers while you haven't actually
   queried yet (with {!query}) or whether you have changed the query since the
   last {!query} (with {!add_query} or {!reset_queries}).
   *)
   method additionals: rr list
   (**
   A resolver can be reused several time.
   Hence it will return only records returned as additionals of the last time you
   ran {!query}.
   Moreover this function has no meaning before you run {!query} (TODO: raise
   an exception then).
   @return the list of resource records in the {b Additional} section of the DNS
   response.
   @raise Name_error when a queried domain name does not exist.
   @raise Nameservers_failure when all nameservers failed us and we got no
   response (which does not mean there is none. We just don't know).
   @raise Not_queried when you try to get answers while you haven't actually
   queried yet (with {!query}) or whether you have changed the query since the
   last {!query} (with {!add_query} or {!reset_queries}).
   *)

   (** {4:confsection Resolver Configuration} *)

   (* TODO:
       * method set_port
       * method port
       * method self_recursion
       *)
   method nameservers : string list
   (** @return the current list of nameservers which will be contacted in the same
    order as the list. *)
   method set_nameservers : string list -> unit
   (** [set_nameservers ns] sets the list [ns] as the new list of nameservers.
    Name servers will be tried and contacted in the order of the provided list.

    As a special case, if an empty list is given, the system's default list of
    name servers (from /etc/resolv.conf) will be used instead. *)

   method set_nocache: bool -> unit
   method nocache: bool

   method timeout : float
   (** @return the currently set timeout, in seconds, after which a query will fail if no
    response has returned. The default value is 5 seconds.
    *)
   method set_timeout : float -> unit
   (** [set_timeout t] sets a new timeout of [t] seconds.
    If you don't set any, the default value is 5 seconds.
    *)

   method retries: int
   (** @return the number of time the resolver algorithm will retry a same
    request on the same server when the previous one failed. *)
   method set_retries: int -> unit 
   (** [set_retries n] set to [n] the number of time the resolver algorithm will
    retry a same request on a same server when previous one failed.
    
    As a special case, if you set it to a negative value, it will be equivalent
    to [0].
    If you don't set this parameter, the default value is [1]. *)

    method recursive: bool
    (** @return [true] if recursion is desired for the query to be sent. *)
    method set_recursive: bool -> unit
    (** [set_recursive true] makes so that the query will "ask" the name server
     to use its recursive capabilities. *)

    method search_list: string list
    (** @return the current search list for hostname lookup ([man 5
     resolv.conf] if you don't know what is the search list. *)
    method set_search_list: string list -> unit
   (** [set_search ns] sets the list [ns] as the new search list.
    Name servers will be tried and contacted in the order of the provided list.
    *)

    method verbose: out_channel option
    (** @return None if verbose mode is disable or else the output channel where
     the verbose output is written to. *)

    (* TODO: should I not propose a semaphore to allow users to write on the
     * same file? This could be useful for log files. *)
    method set_verbose: out_channel option -> unit
    (** Sets an output channel where you want verbose (debug) function to be
     written to. As a special value, [None] disables the verbose mode.
     
     It is up to you to be sure the file can be written in and has not been
     closed, though in any case, this function cannot crash.
     It is also up to you to close this channel (if necessary) when all no
     output should occurre again.
     *)

    method ansi_capable: bool
    (** @return true if you set the verbose output channel as a channel
     recognizing ANSI escape sequences (like most modern Terminal Emulator.
     *)
    method set_ansi_capable: bool -> unit
    (** [set_ansi_capable true] indicates that the verbose output channel is
     ANSI capable. You might want to set this if the output is a terminal
     emulator in order to color different kind of output for instance, but
     disable it when outputing to a file.
     Default to [false].
     *)

   (** {4 Raw Access to Messages} *)

   (** In most cases, you should not have to touch to these methods.
    If you do, be careful not to "break" the messages (in particular the
    query) as no verification will be done here (contrary to high level
    methods which makes all necessary verifications to keep consistent
    messages). *)

   method raw_response : message option
   (** TODO: should return a list of messages, as the answer can be truncated.
   @return an optional list of messages which are the messages received as a
   response to our query.
   It will return [None] if all name servers timed out the query.
   *)
   method raw_query : message
   (**
    [raw_query] enables to study the generated DNS message which will be sent
    over the wire.

    Aware developers can also directly hack this message in order to exploit
    DNS features non-supported by the ODNS library, for instance.
    This should be done after running {!add_query} but before running {!query}.
    @return the DNS message used as a query.
    *)

end;;

(**
[resolver_from_model model] will return a new {!resolver} with all the same
configuration values as [model] (see section {!Dns.resolver.confsection}).
[model] will be untouched and from this point [model] and the new resolver may
diverge in configuration if you change any.

It is a convenient function when you have a specific configuration (but
different from the default one) that you will be using throughout the whole
program, allowing you to make a configuration model instead of always reseting
values.
@return a new {!resolver} with same configuration as [model]
*)
val resolver_from_model: resolver -> resolver

(** Zone Transfer request

NOTE: this resolver has not been much touched since version 0.1 as the current
maintainer (Jehan) is not that interested into zone transfer for its personal
use. Moreover most nameservers don't allow zone transfers so it is hard to test.
If you have chances to test this, don't hesitate. Patches are welcome as
well to set this code in better shape (if needed). *)
class dns_axfr:
    object
        (** {4 High Level Methods} *)

        method get_lst : message list
        method ls : string -> unit
        (**
        [ls d] adds the domain [d] to be listed in the zone transfer. *)
        method query : unit
       (** Actually query the nameservers (using the internal algorithm looping
        through the list of nameservers, waiting for [timeout] seconds and
        retrying [retries] times when failures occurred). *)


       (** {4 Raw Access to Messages} *)

        method raw_response : message option
       (** TODO: should return a list of messages, as the answer can be truncated.
       @return an optional list of messages which are the messages received as a
       response to our query.
       *)
        method raw_query : message
       (**
        [raw_query] enables to study the generated DNS message which will be sent
        over the wire.

        Aware developers can also directly hack this message in order to exploit
        DNS features non-supported by the ODNS library, for instance.
        This should be done before running {!query}.
        @return the DNS message used as a query.
        *)

       (** {4:conf Resolver Configuration} *)

        method nameservers : string list
       (** @return the current list of nameservers which will be contacted in the same
        order as the list. *)
        method set_nameservers : string list -> unit
       (** [set_nameservers ns] sets the list [ns] as the new list of nameservers.
        This list will be tried and contacted in the order of the provided list.

        As a special case, if an empty list is given, the system's default list of
        name servers (from /etc/resolv.conf) will be used instead. *)

        method timeout : float
       (** @return the currently set timeout, in seconds, after which a query will fail if no
        response has returned. The default value is 5 seconds.
        *)
        method set_timeout : float -> unit
       (** [set_timeout t] sets a new timeout of [t] seconds.
        If you don't set any, the default value is 5 seconds.
        *)

        method retries: int
        (** @return the number of time the resolver algorithm will retry a same
        request on the same server when the previous one failed.

        NOTE: the retry algorithm has not been implemented in the AXFR resolver.
        this function is useless. *)
        method set_retries: int -> unit 
        (** [set_retries n] set to [n] the number of time the resolver algorithm will
        retry a same request on a same server when previous one failed.
        
        As a special case, if you set it to a negative value, it will be equivalent
        to [0].
        If you don't set this parameter, the default value is [1].

        NOTE: the retry algorithm has not been implemented in the AXFR resolver.
        this function is useless. *)

        method recursive: bool
        method set_recursive: bool -> unit

        method search_list: string list
        method set_search_list: string list -> unit

        method verbose: out_channel option
        method set_verbose: out_channel option -> unit
        method ansi_capable: bool
        method set_ansi_capable: bool -> unit
end;;

(**********************************)
(*** Easy to use query function ***)
(**********************************)

(*val query: string -> string -> string > (rdata list) * (rdata list) * (rdata list)*)

