(**
 dnshelper.ml --- Helpers to get only relevant information in a high level
 fashion.

 This API is part of ODNS library.
                                                                         
 Copyright (C)  2011  Jehan Hysseo <hysseo at zemarmot point net>
                                                                         
 This program is free software; you can redistribute it and/or modify it 
 under the terms of the GNU General Public License version 2 as          
 published by the Free Software Foundation.                              
                                                                         
 This program is distributed in the hope that it will be useful, but 
 WITHOUT ANY WARRANTY; without even the implied warranty of            
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     
 General Public License for more details.                                

*)

(**
This module provides several helper functions with assumptions on the type of
data. In particular queries will be of the Internet (["IN"]) class, and the
timeout or retry values are the one by default.
The goal is to keep these functions very simple to use, though they still do
sometimes complex computations (like SRV). *)

(* Maybe a/aaaa/address_lookup could check whether the domain is actually in IP
 * form and return it directly if it is the case.
 * But if in IPv4 form while running aaaa_lookup, it should make a reverse
 * query, followed by an aaaa to get an IPv6 supposedly equivalent to the given
 * IPv4. And reciproquely.
 * TODO
 *)
let a_lookup ?(configuration = new Dns.resolver) domain =
    let resolver = Dns.resolver_from_model configuration in
    resolver#add_query ~qtype:"A" ~qclass:"IN" ~qname:domain;
    resolver#query;
    let rec valid_answers ans = function
        | [] -> ans
        | a::l ->
                (* XXX: I use the query name instead of 'domain' variable for the
                 * simple reason that it has been processed in the query. So it
                 * allows to not differentiate for instance two same domains but
                 * one ending without the zero length label ('.' at the end). *)
                if a#name = (List.hd resolver#raw_query#questions)#qname
                then
                    valid_answers ((a#rdata)::ans) l
                else
                    valid_answers ans l
    in
    try
        let answers = valid_answers [] (resolver#answers) in
        let rec rr_to_string result = function
            | [] -> result
            | (Dns.A ip)::l ->
                    rr_to_string (ip::result) l
            | _::l -> rr_to_string result l
        in
        (* Should not be useful, but I remove duplicates anyway. *)
        let rec remove_duplicates = function
            | [] -> []
            | ip::[] -> [ip]
            | ip::l ->
                ip::(remove_duplicates (List.filter
                    (fun ip2 -> ip2 != ip) l))
        in remove_duplicates (rr_to_string [] answers)
    with
    (* resolver#answers may return in error. *)
    | _ -> []
;;

let aaaa_lookup ?(configuration = new Dns.resolver) domain =
    let resolver = Dns.resolver_from_model configuration in
    resolver#add_query ~qname:domain ~qtype:"AAAA" ~qclass:"IN";
    resolver#query;
    let rec valid_answers ans = function
        | [] -> ans
        | a::l ->
                if a#name = (List.hd resolver#raw_query#questions)#qname
                then
                    valid_answers ((a#rdata)::ans) l
                else
                    valid_answers ans l
    in
    try
        let answers = valid_answers [] (resolver#answers) in
        let rec rr_to_string result = function
            | [] -> result
            | (Dns.AAAA ip)::l ->
                    rr_to_string (ip::result) l
            | _::l -> rr_to_string result l
        in
        (* Should not be useful, but I remove duplicates anyway. *)
        let rec remove_duplicates = function
            | [] -> []
            | ip::[] -> [ip]
            | ip::l ->
                ip::(remove_duplicates (List.filter
                    (fun ip2 -> ip2 != ip) l))
        in remove_duplicates (rr_to_string [] (answers))
    with
    (* resolver#answers may return in error. *)
    | _ -> []
;;

let address_lookup ?(configuration = new Dns.resolver) domain =
    let addresses = ref [] in
    let a_lock = Mutex.create () in
    let update_a () =
        let ads = a_lookup ~configuration domain
        in
        Mutex.lock a_lock;
        addresses := !addresses@ads;
        Mutex.unlock a_lock
    in
    let update_aaaa () =
        let ads = aaaa_lookup ~configuration domain
        in
        Mutex.lock a_lock;
        addresses := !addresses@ads;
        Mutex.unlock a_lock
    in
    let a_thread = Thread.create update_a () in
    let aaaa_thread = Thread.create update_aaaa () in
    Thread.join a_thread;
    Thread.join aaaa_thread;
    !addresses
;;

(**************************************
 * These are private functions to be used to order SRV records.
 * I put them out of the main function to get a clearer code, but they should
 * not be used for something else (assumptions are made, hence some checks are
 * missing, making them unsuitable for generic use).
 * Also obviously don't put them in the signature.
 **************************************)

Random.self_init ();;

(* I assume there are already all of same priority. *)
let weight_sum l =
    List.fold_left (fun w1 (_, w2, _, _)-> w1 + w2)
        0 l
;;

let rec finale_reorder l =
    let w_sum = weight_sum l in
    let rand = Random.int (w_sum + 1) in
    let rec get_next running_sum new_list = function
        | [] -> failwith "\nERROR: in function reorder, this error is never supposed to \
        happen! If it does, contact ODNS maintainer.\n"
        | (pr, w, po, t)::l when running_sum + w > rand ->
                (pr, w, po, t), ((List.rev new_list)@l)
        | (pr, w, po, t)::l ->
                get_next (running_sum + w) ((pr, w, po, t)::new_list) l
    in
    match l with
    | [] -> []
    | srv::[] -> [srv]
    | l ->
            let next_srv, remaining_list =
                get_next 0 [] l
            in
            (* TODO: this is more beautiful like this. I could make it recursive
             * terminal, which is more efficient, but do I need efficiency on
             * SRV? Even bigg server won't have more than a dozen of SRV
             * records, right? *)
            next_srv::(finale_reorder remaining_list)
;;

(* Take a list, assumed with same priority (won't check) and place all zero-
 * weighted first, as specified in SVR rfc. *)
(* XXX: another order method could be to sort with a comparison function which
 * always make 0 being inferior to other numbers, but all other numbers are
 * "equal" between each other (= I don't care of orders, just zero first). *)
let order_zero_first prio_list =
    let zero_weight_list, non_zero_weight_list =
        List.partition (fun (_, w, _, _) -> w = 0) prio_list in
    zero_weight_list@non_zero_weight_list
;;

(* take a list of (prio, weight, port, target) and return a list of list, each
 * list having all items with same priority (and ordered in the global list by
 * priority ascendant), beginning with those of weight 0, followed randomly by the
 * others. *)
let reorder_by_prio initial_list =
    let bigger_prio_first_list = List.fast_sort
        (fun (p1, _, _, _) (p2, _, _, _) -> p2 - p1) initial_list in
    let rec gather_by_priority current = function
        | [] -> current
        | (pri, _, _, _)::_ as l ->
                let same_prio_list, diff_prio_list =
                    List.partition (fun (pri2, _, _, _) -> pri = pri2) l in
                gather_by_priority ((order_zero_first same_prio_list)::current) diff_prio_list
    in gather_by_priority [] bigger_prio_first_list
;;

(* There should not be any duplicate, but just in case. *)
let rec remove_duplicates = function
    | [] -> []
    | srv::[] -> [srv]
    | (ip, port)::l ->
            (ip, port)::(remove_duplicates (List.filter
                (fun (ip2, port2) -> ip2 != ip || port2 != port) l))
;;

(* Return the list of (domain name, IPv4 or IPv6 address) in the additionals
 * section of a query. *)
let get_additional_addresses query =
    try
        let additionals = query#additionals in
        let rec get_list current = function
            | [] -> current
            | record::l ->
                    begin
                        match record#rdata with
                        | Dns.A addr | Dns.AAAA addr ->
                                get_list ((record#name, addr)::current) l
                        | _ ->
                                get_list current l
                    end
        in get_list [] additionals
    with
    | _ -> []
    (* Exceptions can occure with nameservers issues or wrong domain name for
     * instance, but here I don't care the reason, I just return a null list. *)
;;

(*********************************************
 * End of private functions,
 * now the public one.
 ********************************************)

(**
 * @param query is the query in the srv fashion, which is
 * _request._protocol.domain_name.
 * @return a list of (IP, port) in an order randomly generated according to SRV rules.
 * In other words, software should not reorder this list and try to connect to
 * the server in the given order.
 *)
(* TODO: maybe use rather protocol/service/target parameters? *)
let srv_lookup ?(configuration = new Dns.resolver) query =
    let resolver = Dns.resolver_from_model configuration in
    resolver#add_query ~qname:query ~qtype:"SRV" ~qclass:"IN";
    resolver#query;
    (* I check that the answer's query is the same one I set. *)
    let rec valid_answers ans = function
        | [] -> ans
        | a::l ->
                if a#name = (List.hd resolver#raw_query#questions)#qname
                then
                    valid_answers ((a#rdata)::ans) l
                else
                    valid_answers ans l
    in
    let answers =
        try
            valid_answers [] (resolver#answers)
        with
        | _ -> []
        (* any of the error, I say simply there is no record. *)
    in
    let rec get_list current = function
        | [] -> current
        | (Dns.SRV (_, _, _, "."))::l ->
                (* I pass such a SRV. *)
                get_list current l
        | (Dns.SRV (pri, weight, port, target))::l ->
                get_list ((pri, weight, port, target)::current) l
        | _::l -> get_list current l
    in
    let additionals = get_additional_addresses resolver in
    match answers with
    | [] ->
            let srv_query = Str.regexp "^_\\([^.]+\\)\\._\\([^.]+\\)\\.\\(.+\\)$" in
            if Str.string_match srv_query query 0
            then
                begin
                    let service = Str.matched_group 1 query in
                    let protocol = Str.matched_group 2 query in
                    let target = Str.matched_group 3 query in
                    let _, addresses_from_add = List.split
                        (List.filter (fun (dn, _) -> dn = target) additionals) in
                    let addresses =
                        match addresses_from_add with
                        | [] ->
                                (* If the A or AAAA records were not sent in
                                * the additionals section in the same time
                                * as the SRV, I explicitely request them. *)
                                address_lookup ~configuration target
                        | ad -> ad 
                    in
                    let port =
                        try 
                            Iana.get_port ~transport:protocol ~service
                        with
                        | Not_found ->
                                (* Can happen if the requested service is not a
                                 * known service, hence the port is unknown.
                                 * In such case, my implementation choice is to
                                 * return a port of -1 instead of an empty
                                 * result. This enables a user to make queries
                                 * with unsupported service as long as one knows
                                 * the default port. *)
                                -1
                    in
                    List.map (fun ad -> (ad, port)) addresses
                end
            else
                (* Wrong syntax *)
                []
    | [Dns.SRV (_, _, _, ".")] -> []
    (* A Target of ".", and nothing else, means that the service is decidedly
     * not available at this domain. *)
    | answers ->
            let processed =
                List.map finale_reorder 
                    (reorder_by_prio (get_list [] answers)) in
            let targets = List.map (fun (_, _, port, target) -> (target, port))
                (List.flatten processed) in
            let addresses = ref [] in
            let addresses_lock = Mutex.create () in
            let update_addresses (dom, port) =
                let ips = address_lookup ~configuration dom in
                let rec aux current = function
                    | [] -> current
                    | ip::l -> aux ((ip, port)::current) l
                in Mutex.lock addresses_lock;
                addresses := aux (!addresses) ips;
                Mutex.unlock addresses_lock
            in
            let rec to_addresses threads = function
                | [] ->
                        let rec wait_all = function
                            | [] -> ()
                            | t::l ->
                                    Thread.join t;
                                    wait_all l
                        in wait_all threads
                | (target, port)::l ->
                        let _, addresses_from_add = List.split
                            (List.filter (fun (dn, _) -> dn = target) additionals) in
                        (*let addresses =*)
                        match addresses_from_add with
                        | [] ->
                                (* If the A or AAAA records were not sent in
                                * the additionals section in the same time
                                * as the SRV, I explicitely request them. *)
                                let t = Thread.create update_addresses (target, port)
                                in to_addresses (t::threads) l
                        | ad ->
                                let rec aux current = function
                                    | [] -> current
                                    | ip::l -> aux ((ip, port)::current) l
                                in
                                Mutex.lock addresses_lock;
                                addresses := aux (!addresses) ad;
                                Mutex.unlock addresses_lock;
                                to_addresses threads l
            in
            to_addresses [] targets;
            remove_duplicates !addresses
;;

(*let mx_lookup domain = *)

