(**
 dnsprint.ml --- Pretty print functions for DNS objects.
                                                                         
 Copyright (C) 2002  Guillaume Valadon <guillaume@valadon.net>
                     Manu Letellier <manu.letellier@gmail.com>
               2010-2011  Jehan Hysseo <hysseo at zemarmot point net>
                                                                         
 This program is free software; you can redistribute it and/or modify it 
 under the terms of the GNU General Public License version 2 as          
 published by the Free Software Foundation.                              
                                                                         
 This program is distributed in the hope that it will be useful, but 
 WITHOUT ANY WARRANTY; without even the implied warranty of            
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     
 General Public License for more details.                                

*)

(** Printing functions for DNS responses. *)

(** [pretty_rdata record] will generate a contextual string translating a DNS
 * resource records in natural language, if possible.
 * @param record is a [Dns.rr] to translate in natural language.
 * @return a string which can be displayed to human being.
 *)
let pretty_record record =
    let name = record#name in
    match record#rdata with
    | Dns.A addr -> 
            Printf.sprintf "%s has IPv4 address %s\n" name  addr
    | Dns.NS addr ->
            Printf.sprintf "%s name server is %s\n" name addr
    | Dns.MD madname ->
            Printf.sprintf "%s has a delivery mail agent in %s\n" name madname
    | Dns.MF madname ->
            Printf.sprintf "%s has a forwarding mail agent in %s\n" name madname
    | Dns.CNAME addr ->
            Printf.sprintf "%s is a alias for %s\n" name addr
    | Dns.SOA (m, r, s, ref, ret, ex, min) ->
            Printf.sprintf "%s start of authority (\n" name
            ^ Printf.sprintf "%32s ;primary source of data for this zone\n" m
            ^ Printf.sprintf "%32s ;responsible's mailbox\n" r
            ^ Printf.sprintf "%32u ;serial (version)\n" s
            ^ Printf.sprintf "%32u ;refresh period\n" ref
            ^ Printf.sprintf "%32u ;retry refresh this often\n" ret
            ^ Printf.sprintf "%32u ;expiration period\n" ex
            ^ Printf.sprintf "%32u ;minimum TTL\n" min
            ^ Printf.sprintf "%28s\n" ")"
    | Dns.MB madname ->
            Printf.sprintf "%s has a mailbox in %s\n" name madname
    | Dns.MG madname ->
            Printf.sprintf "%s's mail group has %s for member\n" name madname
    | Dns.MR madname ->
            Printf.sprintf "%s's mailbox has been renamed into %s\n" name madname
    | Dns.NULL anything ->
            Printf.sprintf "%s NULL record: %s\n" name anything
    | Dns.WKS (addr, proto_num, ports) ->
        (
        try
            let proto =
                Iana.get_protocol_name proto_num in
            let first_port = List.hd ports in
            let first_port_string =
                string_of_int first_port
                ^
                (
                try
                    " ["
                    ^ Iana.get_service ~transport:proto ~port:first_port
                    ^ "]"
                with
                | Not_found -> ""
                )
            in
            let remaining_ports = List.tl ports in
            Printf.sprintf "%s has well known services on %s for %s (%d) on the \
                following ports:\n\t"
                name addr proto proto_num
           ^ List.fold_left
                (fun p1 p2 ->
                    try
                        Printf.sprintf "%s, %s (%d)" p1
                            (Iana.get_service ~transport:proto ~port:p2) p2
                    with
                    | Not_found ->
                        Printf.sprintf "%s, [unknown service on %d]" p1 p2
                )
                first_port_string remaining_ports
           ^ "\n"
        with
        | Not_found ->
                Printf.sprintf "%s has no well known services for the unknown \
                protocol %d\n"
                name proto_num;
        | Failure "hd" | Failure "tl" ->
            (* The bitmap is full of 0. *)
            (
            try
                Printf.sprintf "%s has no well known services on %s for %s\n"
                name addr (Iana.get_protocol_name proto_num);
            with
            | Not_found ->
                Printf.sprintf "%s has no well known services on %s for protocol %d\n"
                name addr proto_num;
            )
        )
    | Dns.PTR ptrdname ->
            Printf.sprintf "%s domain name pointer %s" name ptrdname
    | Dns.HINFO (os, cpu) ->
            Printf.sprintf "%s host information:\n\tCPU = %s\n\tOS= %s\n"
                name cpu os
    | Dns.MINFO (rmailbx, emailbx) ->
            Printf.sprintf "%s mailbox information:\n\t\
            Responsible Mailbox: %s\n\t\
            Error Mailbox: %s\n"
            name rmailbx emailbx
    | Dns.MX (prf, exchange) ->
            Printf.sprintf "%s mail is handled (preference = %d) by %s\n"
            name prf exchange
    | Dns.TXT strs ->
            Printf.sprintf "%s descriptive text:\n" name
            ^ List.fold_left
                (fun before str -> Printf.sprintf "%s\t%s\n" before str)
                    "" strs
    | Dns.SRV (prio, weight, port, target) ->
            let srv_query = Str.regexp "^_\\([^.]+\\)\\._\\([^.]+\\)\\.\\(.+\\)$" in
            begin
                if Str.string_match srv_query name 0
                then
                    let dn = Str.matched_group 3 name in
                    let srv = Str.matched_group 1 name in
                    let prot = Str.matched_group 2 name in
                    Printf.sprintf "%s has a %s service record on %s:"
                        dn srv prot 
                else
                    (* the query seems wrong, but I got a record, so I still display
                     * it. *)
                    Printf.sprintf "%s has a service record:" name
            end
            ^ Printf.sprintf "\n\tpriority \ = %d" prio
            ^ Printf.sprintf "\n\tweight = %d\n\tport = %d" weight port
            ^ Printf.sprintf "\n\ttarget = %s\n" target
    | Dns.AAAA addr -> 
            Printf.sprintf "%s has IPv6 address %s\n" name addr
    | Dns.Unsupported_RDATA (t, data) ->
            Printf.sprintf "Not supported record of type '%d'" t
            ^ Printf.sprintf " and class '%d' discarded\n" record#_class
;;

(** Get a {i string} that represents the ressource record type from 
    an {i integer}. *)
let type_to_string typ = 
   match typ with
   | 1 -> "A"
   | 2 -> "NS"
   | 3 -> "MD"
   | 4 -> "MF"
   | 5 -> "CNAME"
   | 6 -> "SOA"
   | 7 -> "MB"
   | 8 -> "MG"
   | 9 -> "MR"
   | 10 -> "NULL"
   | 11 -> "WKS"
   | 12 -> "PTR"
   | 13 -> "HINFO"
   | 14 -> "MINFO"
   | 15 -> "MX"
   | 16 -> "TXT"
   | 28 -> "AAAA"
   | 33 -> "SRV"
   | 255 -> "ANY"
   | a -> raise (Dns.Unknown_type (string_of_int a))
;;

let class_to_string = function
    | 1 -> "IN"
    | 2 -> "CS"
    | 3 -> "CH"
    | 4 -> "HS"
    | 255 -> "*"
    | t -> raise (Dns.Unknown_class (string_of_int t));;

(** Print the content of dns header flags. *)
let pretty_flags integer =
    (* QR bit *)
    if (integer land 0x8000) == 0x8000
    then
        "\tResponse message to a "
    else
        "\tQuery message of kind "
    (* OPCODE 4 bits *)
    ^ begin
        match (integer land 0x7800) lsr 11 with
        | 0 -> "standard query\n"
        | 1 -> "inverse query\n"
        | 2 -> "server status request\n"
        | _ -> "unknown query (reserved for future use)\n"
    end
    (* AA bit *)
    (* TODO: valid bit only for responses. *)
    ^ 
    if (integer land 0x0400) == 0x0400
        then
            "\tNameserver is authoritative\n"
    else
        "\tNameserver is not authoritative\n"
    (* TC bit *)
    ^
    if (integer land 0x0200) == 0x0200
    then
        "\tMessage was truncated\n"
    else
        "\tMessage was not truncated\n"
    (* RD bit *)
    (* this bit may be set in a query and
     * is copied into the response *)
    ^
    if (integer land 0x0100) == 0x0100
    then
        "Recursion is desired\n"
    else
        "Recursion is not desired\n"
    (* RA bit *)
    (* TODO: Only in a response. *)
    ^
    if (integer land 0x0080) == 0x0080
    then
        "\tRecursion is available\n"
    else
        "\tRecursion is not available\n"
    (* Z bits *)
    ^
    if (integer land 0x0070) == 0
    then
        "Z bits are 0\n"
    else
        "Z bits are not 0\n"
    (* RCODE bits *)
    ^
    (match (integer land 0x000F) with
    | 0 -> "\tNo error\n"
    | 1 -> "\tFormat error - The name server was \ 
        unable to interpret the query\n"
    | 2 -> "\tServer failure - The name server was \
        unable to process this query due to a problem with the \
        name server\n"
    | 3 -> "\tName Error - Meaningful only for \
        responses from an authoritative name server, this \
        code signifies that the domain name referenced in \
        the query does not exist.\n"
    | 4 -> "\tNot Implemented - The name server \
        does not support the renameed kind of query.\n"
    | 5 -> "\tRefused - The name server refuses \
        to perform the specified operation for policy \
        reasons.  For example, a name server may not wish \
        to provide the information to the particular renameer, \
        or a name server may not wish to perform a particular 
        operation (e.g., zone transfer) for particular data.\n"
    | _ -> "\tReserved for future use.\n")
;;

(** Print a {i Dns.dns_data_header} object. *)
let pretty_header header =
    "==== HEADER ====\n"
    ^ ("ID: " ^ (string_of_int header#id) ^ "\n")
    ^ ("Flags: " ^ (string_of_int header#flags) ^ "\n")
    ^ pretty_flags header#flags ^ "\n"
    ^ ("Number of questions: " ^ (string_of_int header#qdcount) ^"\n")
    ^ ("Number of answers: " ^ (string_of_int header#ancount) ^ "\n")
    ^ ("Number of authority records: " ^ (string_of_int header#nscount) ^ "\n")
    ^ ("Number of additionnal records: " ^ (string_of_int header#arcount) ^ "\n")
    ^ "================\n";
;;

(** Print a {i Dns.dns_data_nameion list}. *)
let rec pretty_queries liste =
    match liste with
    | query::l ->
            "Queried Domain Name: " ^ query#qname ^ "\n"
            ^ "Query Type: "^ type_to_string query#qtype ^ "\n"
            ^ "Class: " ^ string_of_int query#qclass ^ "\n\n"
            ^ pretty_queries l
    | [] -> ""
;;

(** Print a {i Dns.dns_data_resonse list}. *)
let rec pretty_rrs liste =
    match liste with
    | answer::l ->
            "Resource Domain Name: " ^ (answer#name) ^ "\n"
            ^ "Resource Record Type: " ^ (type_to_string answer#_type)^"\n"
            ^ "Class: " ^ (string_of_int answer#_class)^"\n"
            ^ "Time To Live (caching time before discard): "
            ^ (string_of_int answer#ttl) ^ "\n"
            ^ "Rdata: "
            (*print_rdata answer#get_rdata;*)
            ^ pretty_rrs l
    |[] -> ""
;;

(** Print a [Dns.message]. *)
let pretty_message message =
   let hd = message#header in
   pretty_header hd
   (*print_flags hd#flags;*)
   (*if hd#qdcount != 0*)
   (*then*)
      ^ pretty_queries message#questions
      (*if hd#ancount != 0*)
      (*then*)
      ^ "==== QUESTIONS ====\n\n"
      ^ pretty_rrs message#answers
      (*if hd#nscount != 0*)
      (*then*)
      ^ "==== AUTHORITIES ====\n\n"
      ^ pretty_rrs message#authorities
      (*if hd#arcount != 0*)
      (*then*)
      ^ "==== ADDITIONALS ====\n\n";
      ^ pretty_rrs message#additionnals
;;
