(**
 host.ml --- "ring", the dig/host-like command in OCaml
                                                                         
 Copyright (C) 2002  Guillaume Valadon <guillaume@valadon.net>
                     Manu Letellier <manu.letellier@gmail.com>
               2010-2011  Jehan Hysseo <hysseo at zemarmot point net>
                                                                         
 This program is free software; you can redistribute it and/or modify it 
 under the terms of the GNU General Public License version 2 as          
 published by the Free Software Foundation.                              
                                                                         
 This program is distributed in the hope that it will be useful, but 
 WITHOUT ANY WARRANTY; without even the implied warranty of            
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     
 General Public License for more details.                                

*)

(* type that represents 'host' configuration *)
type config =
{
   typ : string;
   hostserv : string list;
   cls : string;
   timeout: float;
   retries: int;
   recurs : bool ref;
   lst : bool ref;
   verbose : bool ref;
   noterm : bool ref;
   nocache: bool ref;
}
;;

(* This function will output answer the 'host' way *)
(* TODO: must accept several queries (though mainly unused, in theory, there can
 * be several queries in the same message)
 * TODO: this function will have to be deleted once I have time to look at axfr. *)
let print_query_answer answ qtype host =
    let answered = ref false in
    let rec aux answers =
       match answers with
       | a::l -> 
          (if qtype = 255 (* all records *)|| a#_type = qtype
          (* TODO actually some kind of query type can imply several significant
           * response types. *)
          then
              answered := true;
              print_string (Dnsprint.pretty_record a)
              );
              aux l
       | [] ->
               if not !answered
               then
                   Printf.printf "%s has no %s record\n" host
                       (String.uppercase (Dnsprint.type_to_string qtype))
    in aux answ
;;


let usage =
   "ring usage: ring [--help] [-nlfT] [-r retries] [-w timeout] [-t type] [-c class] host [server]...\n";;

(* Let's parse the command line *)
let parse_args = 
    let default = ref
    {
        typ = "*";
        hostserv = [];
        cls = "IN";
        recurs = ref false;
        lst = ref false;
        verbose = ref false;
        noterm = ref false;
        timeout = 3.;
        retries = 1;
        nocache = ref false;
    } 
   and set_type def s =
       def := { !def with
       typ = String.uppercase s}
   and set_cls def x =
       def := { !def with cls = x}
   and set_timeout def x =
       def := { !def with timeout = x}
   and set_retries def x =
       def := { !def with retries = x}
   and anon def s =
       def := { !def with hostserv = !def.hostserv@[s]}
              in
              let lst = 
                  [("-t", Arg.String (set_type default), "specifies the query\
                    type (default to any: '*')");
                  ("-c", Arg.String (set_cls default), "specifies the query\
                    class (default to Internet data: 'IN')");
                  ("-T", Arg.Set !default.noterm, "turns off 'Terminal mode'. \
                  By default output will be colorized, but you might want \ 
                  to drop the output into some file. In this case, it is \
                  better to turn this feature off.");
                  ("-f", Arg.Set !default.nocache, "deactivate cache support. \
                  Force queries to run against nameserver instead of checking \
                  cache results.\ 
                  Results won't be saved in cache (default: activated).");
                  ("-l", Arg.Set !default.lst, "turns on 'list mode'");
                  ("-v", Arg.Set !default.verbose, "turns on 'verbose mode' \
                  (will print name server's IPs and tries, errors, additionals \
                  and authorities records)");
                  ("-n", Arg.Set !default.recurs, "disables recursive query");
                  ("-r", Arg.Int (set_retries default), "specifies the number \
                        of retries a single query will be made in case of \
                        failure (default to 1 retry). A negative value is \
                        equivalent to 0 (no retry)");
                  ("-w", Arg.Float (set_timeout default), "specifies the timeout \
                        in seconds (default to 3 seconds). A negative value \
                        means there is no timeout (wait forever a reply)");
                  ]
    in
    Arg.parse lst (anon default) usage;
    !default
;;


let rec classic nameservers host conf search =
    let print_section section =
        if (not !(conf.noterm))
        then
            begin
                let esc = "^[0;7m" in
                esc.[0] <- Char.chr 27;
                print_string esc
            end;
        Printf.printf "\n= %s SECTION =\n" section;
        if (not !(conf.noterm))
        then
            begin
                let esc = "^[0m" in
                esc.[0] <- Char.chr 27;
                print_string esc
            end
    in
    let dns = new Dns.resolver in
    dns#set_nameservers nameservers;

    (* -n : cancel recursive request ?  *)
    dns#set_recursive (not !(conf.recurs));
    (* -w: set timeout *)
    dns#set_timeout conf.timeout;
    (* -r: number of retries. *)
    dns#set_retries conf.retries;
    (* -v: verbose mode *)
    if !(conf.verbose)
    then
        dns#set_verbose (Some stdout);

    dns#set_ansi_capable (not !(conf.noterm));
    dns#set_nocache !(conf.nocache);

    (* XXX: this option has already been used! If it is true, it would have
     * run axfr, not classic! *)
    (*if !(conf.lst)*)
    (*then*)
    (*dns#add_query ~qname:host ~qtype:"*" ~qclass:conf.cls*)
    (*else*)

    (* Build the query using -c and -t *)
    dns#add_query ~qname:host ~qtype:conf.typ ~qclass:conf.cls;

    (* Send the query. *)
    dns#query;

    (* finally print the answer *)
    try
        let answers = dns#answers in
        if !(conf.verbose) && answers != []
        then
            print_section "ANSWER";
        if List.length answers != 0
        then
            List.iter (fun rr -> print_string (Dnsprint.pretty_record rr))
                answers
        else
            Printf.printf "%s has no %s records of class %s.\n"
                host conf.typ conf.cls;

        (* In verbose mode, I print also the additionals and authorities. *)
        if !(conf.verbose)
        then
            (
                (
                match dns#additionals with
                | [] -> ()
                | add ->
                    print_section "ADDITIONAL";
                    List.iter (fun rr -> print_string (Dnsprint.pretty_record rr))
                    add
                );
                (
                match dns#authorities with
                | [] -> ()
                | auth ->
                    print_section "AUTHORITIES";
                    List.iter (fun rr -> print_string (Dnsprint.pretty_record rr))
                    auth
                );
            )
    with
    | Dns.Name_error ->
            if search != []
            then
                classic nameservers (host ^ "." ^ (List.hd search)) conf (List.tl search)
            else
                Printf.printf "Host %s not found.\n" host
    | Dns.Nameservers_failure ->
            if not !(conf.verbose)
            then
                Printf.printf "No nameserver could answer the query.\n\
                It usually mean they either timed-out, \
                had internal server error, \
                or refused the query.\n\
            Set verbose mode (-v) to get more detail.\n"
    | Dns.Not_queried ->
            failwith "An unexpected Not_queried error occurred. Contact ring's \
            maintainer.\n"
;;
    (* TODO: rcode = 3:
    | 3 -> 
            if search != []
            then
                classic nameservers (host ^ "." ^ (List.hd search)) conf (List.tl search)
            else
                Printf.printf "Host %s not found.\n" host
*)

let axfr nameservers host =
    let axfr = new Dns.dns_axfr in
    axfr#set_nameservers nameservers;
   axfr#ls host;
   axfr#query;
   let lst = axfr#get_lst in
   match (List.hd lst)#header#rcode with
   | 0 -> 
      List.iter (fun x ->
	 print_query_answer x#answers 255 host) lst; 
   | 5 -> 
      print_string "Server failed: Query refused\n"
   | other -> 
      print_string ("TODO: error number "^(string_of_int other)^".")
;;

try
    let conf = parse_args in
    let host = List.hd conf.hostserv in
    let rec keep_uniques = function
        | [] -> []
        | a::[] -> [a]
        | a::l ->
                a::(keep_uniques (List.filter (fun x -> x != a) l))
    in
    let nameservers =
        if List.length conf.hostserv = 2
        then
            begin
                let srv = List.nth conf.hostserv 1 in
                Printf.printf "Using domain name server: %s\n" srv;
                let ips = Tools.getaddr srv in
                Printf.printf "Name server's addresses: %s\n\n"
                    (List.fold_left (fun s1 s2 -> s1 ^ ", " ^ s2)
                    (List.hd ips) (List.tl ips));
                flush stdout;
                ips
            end
        else if List.length conf.hostserv > 2
        then
            begin
                let srvs = List.tl conf.hostserv in
                Printf.printf "Using domain name servers in this order: %s\n"
                    (List.fold_left (fun s1 s2 -> s1 ^ ", " ^ s2)
                        (List.hd srvs) (List.tl srvs));
                let ips = keep_uniques
                    (List.flatten (List.map Tools.getaddr srvs)) in
                (* XXX: Normally Tools.getaddr will anyway return a
                 * Invalid_IP_or_DN exception if there is no IP!
                 * So I should not have to care or ips being empty
                 * (is it sure?). *)
                Printf.printf "Name servers' addresses: %s\n\n"
                    (List.fold_left (fun s1 s2 -> s1 ^ ", " ^ s2)
                    (List.hd ips) (List.tl ips));
                flush stdout;
                ips
            end
        else
            []
    in
    (* Let's use command line arguments *)
    if !(conf.lst)
    then
        axfr nameservers host
    else
        (* TODO: add again search in a better way. *)
        classic nameservers host conf [] (*search*)
with
| Failure "hd" ->
    print_string usage;
    exit 1
| Dns.Unknown_type str ->
        print_string ("Unknown query type: " ^ str ^ "\n");
        exit 2
| Dns.Unknown_class str ->
        print_string ("Unknown query class: " ^ str ^ "\n");
        exit 2
| Tools.Invalid_IP_or_DN dn ->
        Printf.printf "The provided name server (%s) is not a valid IP \
        or domain name.\n" dn;
        exit 1
;;

(*let a = Dns_helper.address_lookup "zemarmot.net" in
List.iter (fun a -> print_string (a ^ "\n")) a;;
let s = Dns_helper.srv_lookup "_xmpp-client._tcp.zemarmot.net" in
List.iter (fun (a, ip) -> Printf.printf "%s %d\n" a ip) s;;*)
