(**
 iana.ml --- Internet Assigned Numbers Authority support library
                                                                         
 Copyright (C)  2011  Jehan Hysseo <hysseo at zemarmot point net>
                                                                         
 This program is free software; you can redistribute it and/or modify it 
 under the terms of the GNU General Public License version 2 as          
 published by the Free Software Foundation.                              
                                                                         
 This program is distributed in the hope that it will be useful, but 
 WITHOUT ANY WARRANTY; without even the implied warranty of            
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     
 General Public License for more details.                                

*)

type protocol =
    | TCP
    | UDP
;;

(********************************************************)
(************* PROTOCOLS ASSIGNEMENT ********************)
(********************************************************)

(* Parse a /etc/protocols style file (man 5 protocols) and returns a list of
 * (protocol name or alias, protocol number).
 *)
let load_protocol_numbers file =
    let blanks = Str.regexp "[ \t]+" in
    try
        let f = open_in file in
        let rec process_line numbers =  
            try
                let line = input_line f in
                let line =
                    try
                        (* I remove any comment. *)
                        String.sub line 0 (String.index line '#')
                    with
                    | Not_found -> line
                in
                match Str.split_delim blanks line with
                | [] | ""::_ ->
                        (* A white line.
                         * Or a line beginning by blank.
                         *)
                        process_line numbers
                | protocol::number::aliases ->
                        begin try
                            let aliases = List.map
                                (fun alias -> (alias, int_of_string number))
                                aliases
                            in
                            (* XXX: it is important that the official name is
                             * the first in the list so that this is the one
                             * returned when requested by number. *)
                            process_line ((protocol, int_of_string
                                number)::aliases@numbers)
                        with
                        | Failure "int_of_string" ->
                                prerr_string "Invalid protocol number in the \
                                protocols file\n";
                                process_line numbers
                        end
                | _ ->
                        prerr_string ("Unrecognized pattern in the \
                            IANA's protocol-numbers file:\n" ^ line ^ "\n");
                            process_line numbers
            with
            | End_of_file ->
                    close_in_noerr f;
                    numbers
        in
        process_line []
    with
    | Sys_error _ ->
            (* The file could not be opened. *)
            []
;;

let assigned_protocols =
    lazy (load_protocol_numbers "/etc/protocols");;

(* Public Interface. *)

let get_protocol_number ~name =
    let protocols = Lazy.force assigned_protocols in
    List.assoc name protocols
;;

let get_protocol_name ~number =
    let protocols = Lazy.force assigned_protocols in
    fst (List.find (fun (_, n) -> n = number) protocols)
;;

(********************************************************)
(************* PORT ASSIGNEMENT *************************)
(********************************************************)

(* Parse a /etc/services style file (man 5 services) and returns a list of
 * ((service name or alias, transport protocol), port).
 *)
let load_port_numbers file =
    let blanks = Str.regexp "[ \t]+" in
    (* For compatibility reasons, the port and protocols can be separated either
     * by '/' or ',', though the comma is deprecated. *)
    let port_proto_regexp = Str.regexp "^\\([0-9]+\\)[/,]\\(.+\\)$" in
    try
        let f = open_in file in
        let rec process_line ports =  
            try
                let line = input_line f in
                let line =
                    try
                        (* I remove any comment. *)
                        String.sub line 0 (String.index line '#')
                    with
                    | Not_found -> line
                in
                match Str.split_delim blanks line with
                | [] | ""::_ ->
                        (* A white line.
                         * A port with no associated keyword. For now I
                         * don't need such a port, so I don't save this.
                         *)
                        process_line ports
                | keyword::port_proto::aliases ->
                        begin try
                            if (Str.string_match port_proto_regexp port_proto 0)
                            then
                                let proto = Str.matched_group 2 port_proto in
                                let port =
                                    int_of_string (Str.matched_group 1 port_proto) in
                                let aliases = List.map
                                    (fun alias -> ((alias, proto), port)) aliases in
                                process_line
                                (((keyword, proto), port)::aliases@ports)
                            else
                                begin
                                    prerr_string ("Unrecognized protocol in the IANA's port-number file:\n" ^ line ^ "\n");
                                    process_line ports
                                end
                        with
                        | Failure "int_of_string" ->
                                (* Should normally not happen as the
                                 * regexp checked it was numbers. *)
                                process_line ports
                        end
                | _ ->
                        prerr_string ("Unrecognized pattern in the \
                            IANA's port-number file:\n" ^ line ^ "\n");
                        process_line ports
            with
            | End_of_file ->
                    close_in_noerr f;
                    ports
        in
        process_line []
    with
    | Sys_error _ ->
            (* The file could not be opened. *)
            []
;;

(* XXX: the assigned ports file is actually /etc/services.
 * Though it may not be the most updated of all, it should be nevertheless
 * enough in most uses.
 * Of course this file is relevant only for Linux (and probably UNIX or
 * UNIX-lile systems, though it is not verified, but definitely not on Windows
 * system).
 * If someone wants to make this portable, I welcome any information from users
 * of any system where it might be different and I will make conditional parsing
 * and load here with any other equivalent.
 * XXX: I don't want to load this file when it is not useful. Hence the solution
 * is to make it a lazy computation. It will be loaded only when required.
 *)
let assigned_ports =
    lazy (load_port_numbers "/etc/services");;

let get_port ?(transport = "tcp") ~service:keyword =
    let ports = Lazy.force assigned_ports in
    List.assoc (keyword, transport) ports
;;

let get_service ?(transport = "tcp") ~port =
    let ports = Lazy.force assigned_ports in
    fst (fst
        (List.find (fun ((_, prot), po) -> prot = transport && port = po) ports))
;;
