(**
 iana.mli --- Internet Assigned Numbers Authority support library
                                                                         
 Copyright (C)  2011  Jehan Hysseo <hysseo at zemarmot point net>
                                                                         
 This program is free software; you can redistribute it and/or modify it 
 under the terms of the GNU General Public License version 2 as          
 published by the Free Software Foundation.                              
                                                                         
 This program is distributed in the hope that it will be useful, but 
 WITHOUT ANY WARRANTY; without even the implied warranty of            
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     
 General Public License for more details.                                

 @see <http://tools.ietf.org/html/rfc3232> RFC 3232: Assigned Numbers.
 @see <http://www.iana.org> Internet Assigned Numbers authority.
 @see '/etc/services' The network services file under Linux and Unix systems.
 @see '/etc/protocols' The assigned IP protocols file under Linux and Unix systems.

*)

(** This module provides an interface to default "number" assignement on the
 Internet, for instance ports assignement on a specific protocol, or numbers
 assigned to Internet protocols, as defined by the
 {{:http://www.iana.org}Internet Assigned Numbers Authority}.
 *)

(**
[get_protocol_number ~name] returns the official IANA registered number for the
protocol which has for official name or alias [name].
@param name the official name or a widespread alias of the protocol.
@return the registered number if it exists.
@raise Not_found when there is no protocol of this name or alias.
@see <http://www.iana.org/assignments/protocol-numbers/protocol-numbers.xml> The
most up-to-date version of protocol numbers as assigned by IANA.
@see '/etc/protocols' The assigned IP protocols file under Linux and Unix systems.
 *)
val get_protocol_number: name:string -> int

(**
[get_protocol_name ~number:n] returns the official name for the
protocol which has for registered number [n].
@param number the IANA registered number.
@return the official protocol name.
@raise Not_found when there is no protocol registered for this number.
@see <http://www.iana.org/assignments/protocol-numbers/protocol-numbers.xml> The
most up-to-date version of protocol numbers as assigned by IANA.
@see '/etc/protocols' The assigned IP protocols file under Linux and Unix systems.
 *)
val get_protocol_name: number:int -> string

(*val load_port_numbers: string -> (string * int) list * (string * int) list*)

(**
[get_port ~transport:proto ~service:keyword]  returns the usual port used for the
service whose keyword is [keyword] with the protocol [proto], as assigned by the
IANA.
@return the assigned port, usually used by the given couple (proto/keyword).
@param service the official keyword (or a common alias) of the Application Protocol we are looking-up.
@param transport the Transport Protocol (IP protocol).
Common values are "tcp" and "udp" (default: ["tcp"]).
@raise Not_found when there is no registered service for the couple
(proto/keyword).
@see <http://www.iana.org/assignments/port-numbers> The most up-to-date
version of the IANA assigned ports.
@see '/etc/services' The version used locally on your system.
*)
val get_port: ?transport:string -> service:string -> int

(**
[get_service ~transport ~port]  returns the official service name used which has
the given [port] registration for the protocol [transport] (as assigned by IANA).
@param port port registered by IANA for the researched service.
@param transport the Transport Protocol (IP protocol) associated to the port
as registered by IANA.
Common values are "tcp" and "udp" (default: ["tcp"]).
@return the official service name used on [port].
@raise Not_found when there is no registered service for this port and transport.
@see <http://www.iana.org/assignments/port-numbers> The most up-to-date
version of the IANA assigned ports.
@see '/etc/services' The version used locally on your system.
*)
val get_service: ?transport:string -> port:int -> string


(* TODO: in some future, maybe save also the comment-part of the assigned ports
 * in /etc/services. It can serve as a "human language" name for the service. *)
